/* $Id: l3dss1_nt.c,v 0.0.0.0 0000/00/00 14:19:36 jolly Exp $

nt mode, implemented by jolly

based on the l3dss1.c with the original notice:

 *
 * EURO/DSS1 Network D-channel protocol
 *
 * German 1TR6 D-channel protocol
 *
 * Author       Karsten Keil
 *              based on the teles driver from Jan den Ouden
 * Copyright    by Karsten Keil      <keil@isdn4linux.de>
 * 
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 *
 * For changes and modifications please read
 * ../../../Documentation/isdn/HiSax.cert
 *
 * Thanks to    Jan den Ouden
 *              Fritz Elfert
 *
 */

#define __NO_VERSION__
#include "hisax.h"
#include "isdnl1.h"
#include "isdnl3.h"
#include "l3dss1_nt.h"
#include <linux/ctype.h>
#include <linux/config.h>
#include <linux/rtc.h>

extern char *HiSax_getrev(const char *revision);
const char *dss1_nt_revision = "$Revision: 0.0.0.0 $";

#define EXT_BEARER_CAPS 1

#define	MsgHead(ptr, cref, mty) \
	*ptr++ = 0x8; \
	if (cref == -1) { \
		*ptr++ = 0x0; \
	} else { \
		*ptr++ = 0x1; \
		*ptr++ = cref^0x80; \
	} \
	*ptr++ = mty

/* name of all states */
char *state_name[]= {
	"null_state",
	"call_initiated",
	"overlap_sending",
	"outgoing_call_proceeding",
	"call_delivered",
	"*unused*",
	"call_present",
	"call_received",
	"connect_request",
	"incoming_call_proceeding",
	"active",
	"disconnect_request",
	"disconnect_indication",
	"*unused*",
	"*unused*",
	"suspend_request",
	"*unused*",
	"resume_request",
	"*unused*",
	"release_request",
	"*unused*",
	"*unused*",
	"call_abort",
	"*unused*",
	"*unused*",
	"overlap_receiving",
	"active_after_bc_rel", /* state 26 */
	"null_after_bc_rel", /* state 27 */
} ;

static char 
*l3_state_name(int state)
{
	static char name[64];

	if (state<0 || state>(sizeof(state_name)/sizeof(char *))) {
		sprintf(name,"*out of range*(%d)",state);
		return(name);
	}

	sprintf(name,"%s(%d)",state_name[state],state);
	return(name);
}

static struct l3_process
*getl3proc_nt(struct PStack *st, int cr, int tei)
{
	struct l3_process *p = st->l3.proc;

#ifdef CONFIG_HISAX_MULTIPOINT
	if (!(cr & 0x80)) { /* the tei is from user */
		while(p)
			if (p->callref==cr && (p->dss1_nt.tei==tei || tei==0))
				return (p);
			else
				p = p->next;
		return (NULL);
	}
#endif

	while (p)
		if (p->callref == cr)
			return (p);
		else
			p = p->next;
	return (NULL);
}


/* calculate time elements */

#define FEBRUARY		2
#define STARTOFTIME		1970
#define SECDAY			86400L
#define SECYR			(SECDAY * 365)
#define leapyear(year)		((year) % 4 == 0)
#define days_in_year(a)		(leapyear(a) ? 366 : 365)
#define days_in_month(a)	(month_days[(a) - 1])

static int month_days[12] = {
	31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

static void
l3dss1_nt_to_tm(unsigned long tim, struct rtc_time * tm)
{
	long hms, day;
	int i;

	day = tim / SECDAY;
	hms = tim % SECDAY;

	/* Hours, minutes, seconds are easy */
	tm->tm_hour = hms / 3600;
	tm->tm_min = (hms / 60) % 60;
	tm->tm_sec = hms % 60;

	/* Number of years in days */
	for (i = STARTOFTIME; day >= days_in_year(i); i++)
		day -= days_in_year(i);
	tm->tm_year = i;

	/* Number of months in days left */
	if (leapyear(tm->tm_year))
		days_in_month(FEBRUARY) = 29; /* set leap year */
	for (i = 1; day >= days_in_month(i); i++)
		day -= days_in_month(i);
	days_in_month(FEBRUARY) = 28; /* reset leap year */
	tm->tm_mon = i;

	/* Days are what is left over (+1) from all that. */
	tm->tm_mday = day + 1;

	/*
	 * Determine the day of week
	 */
	day = tim / SECDAY;
	tm->tm_wday = (day + 3) % 7;

	/*
	 * Special daylight saving time calculation
		will be added here
	 */
}


/**********************************************/
/* get a new invoke id for remote operations. */
/* Only a return value != 0 is valid          */
/**********************************************/
static unsigned char nt_new_invoke_id(struct PStack *p)
{
	unsigned char retval;
	unsigned long flags;
	int i;
  
	i = 32; /* maximum search depth */

	save_flags(flags);
	cli();

	retval = p->prot.dss1_nt.last_invoke_id + 1; /* try new id */
	while ((i) && (p->prot.dss1_nt.invoke_used[retval >> 3] == 0xFF)) {
		p->prot.dss1_nt.last_invoke_id = (retval & 0xF8) + 8;
		i--;
	}  
	if (i) {
		while (p->prot.dss1_nt.invoke_used[retval >> 3] & (1 << (retval & 7)))
		retval++; 
	} else
		retval = 0;
	p->prot.dss1_nt.last_invoke_id = retval;
	p->prot.dss1_nt.invoke_used[retval >> 3] |= (1 << (retval & 7));
	restore_flags(flags);

	return(retval);  
} /* nt_new_invoke_id */

/*************************/
/* free a used invoke id */
/*************************/
static void nt_free_invoke_id(struct PStack *p, unsigned char id)
{ unsigned long flags;

  if (!id) return; /* 0 = invalid value */

  save_flags(flags);
  cli();
  p->prot.dss1_nt.invoke_used[id >> 3] &= ~(1 << (id & 7));
  restore_flags(flags);
} /* nt_free_invoke_id */  


/**********************************************************/
/* create a new l3 process and fill in dss1 specific data */
/**********************************************************/
static struct l3_process
*dss1_nt_new_l3_process(struct PStack *st, int cr, int tei)
{  struct l3_process *proc;
   int i;

   if (!(proc = new_l3_process(st, cr))) 
     return(NULL);

   proc->prot.dss1_nt.tei = tei;
   i = 0;
   while(i < MULTIPOINT_NUM) {
   	proc->prot.dss1_nt.teis[i] = -1;
   	proc->prot.dss1_nt.states[i] = 0;
	i++;
   }
	
   proc->prot.dss1_nt.invoke_id = 0;
   proc->prot.dss1_nt.remote_operation = 0;
   proc->prot.dss1_nt.uus1_data[0] = '\0';
   
   return(proc);
} /* dss1_nt_new_l3_process */

/************************************************/
/* free a l3 process and all dss1 specific data */
/************************************************/ 
static void
dss1_nt_release_l3_process(struct l3_process *p)
{
	/* release bchannel */
	if (p->para.bchannel_inuse)
		bchannel_nt_free(p->st->l1.hardware,p->para.bchannel);
	p->para.bchannel_inuse = 0;

   nt_free_invoke_id(p->st,p->prot.dss1_nt.invoke_id);
   release_l3_process(p);
} /* dss1_nt_release_l3_process */
 
/********************************************************/
/* search a process with invoke id id and dummy callref */
/********************************************************/
static struct l3_process *
l3dss1_nt_search_dummy_proc(struct PStack *st, int id)
{ struct l3_process *pc = st->l3.proc; /* start of processes */

  if (!id) return(NULL);

  while (pc)
   { if ((pc->callref == -1) && (pc->prot.dss1_nt.invoke_id == id))
       return(pc);
     pc = pc->next;
   } 
  return(NULL);
} /* l3dss1_nt_search_dummy_proc */

/*******************************************************************/
/* called when a facility message with a dummy callref is received */
/* and a return result is delivered. id specifies the invoke id.   */
/*******************************************************************/ 
static void 
l3dss1_nt_dummy_return_result(struct PStack *st, int id, u_char *p, u_char nlen)
{ isdn_ctrl ic;
  struct IsdnCardState *cs;
  struct l3_process *pc = NULL; 

  if ((pc = l3dss1_nt_search_dummy_proc(st, id)))
   { L3DelTimer(&pc->timer); /* remove timer */

     cs = pc->st->l1.hardware;
     ic.driver = cs->myid;
     ic.command = ISDN_STAT_PROT;
     ic.arg = DSS1_STAT_INVOKE_RES;
     ic.parm.dss1_io.hl_id = pc->prot.dss1_nt.invoke_id;
     ic.parm.dss1_io.ll_id = pc->prot.dss1_nt.ll_id;
     ic.parm.dss1_io.proc = pc->prot.dss1_nt.proc;
     ic.parm.dss1_io.timeout= 0;
     ic.parm.dss1_io.datalen = nlen;
     ic.parm.dss1_io.data = p;
     nt_free_invoke_id(pc->st, pc->prot.dss1_nt.invoke_id);
     pc->prot.dss1_nt.invoke_id = 0; /* reset id */

     cs->iif.statcallb(&ic);
     dss1_nt_release_l3_process(pc); 
   }
  else
   l3_debug(st, "nt dummy return result id=0x%x result len=%d",id,nlen);
} /* l3dss1_nt_dummy_return_result */

/*******************************************************************/
/* called when a facility message with a dummy callref is received */
/* and a return error is delivered. id specifies the invoke id.    */
/*******************************************************************/ 
static void 
l3dss1_nt_dummy_error_return(struct PStack *st, int id, ulong error)
{ isdn_ctrl ic;
  struct IsdnCardState *cs;
  struct l3_process *pc = NULL; 

  if ((pc = l3dss1_nt_search_dummy_proc(st, id)))
   { L3DelTimer(&pc->timer); /* remove timer */

     cs = pc->st->l1.hardware;
     ic.driver = cs->myid;
     ic.command = ISDN_STAT_PROT;
     ic.arg = DSS1_STAT_INVOKE_ERR;
     ic.parm.dss1_io.hl_id = pc->prot.dss1_nt.invoke_id;
     ic.parm.dss1_io.ll_id = pc->prot.dss1_nt.ll_id;
     ic.parm.dss1_io.proc = pc->prot.dss1_nt.proc;
     ic.parm.dss1_io.timeout= error;
     ic.parm.dss1_io.datalen = 0;
     ic.parm.dss1_io.data = NULL;
     nt_free_invoke_id(pc->st, pc->prot.dss1_nt.invoke_id);
     pc->prot.dss1_nt.invoke_id = 0; /* reset id */

     cs->iif.statcallb(&ic);
     dss1_nt_release_l3_process(pc); 
   }
  else
   l3_debug(st, "nt dummy return error id=0x%x error=0x%lx",id,error);
} /* l3dss1_nt_error_return */

/*******************************************************************/
/* called when a facility message with a dummy callref is received */
/* and a invoke is delivered. id specifies the invoke id.          */
/*******************************************************************/ 
static void 
l3dss1_nt_dummy_invoke(struct PStack *st, int cr, int id, 
                    int ident, u_char *p, u_char nlen)
{ isdn_ctrl ic;
  struct IsdnCardState *cs;
  
  l3_debug(st, "nt dummy invoke %s id=0x%x ident=0x%x datalen=%d",
               (cr == -1) ? "local" : "broadcast",id,ident,nlen);
  if (cr >= -1) return; /* ignore local data */

  cs = st->l1.hardware;
  ic.driver = cs->myid;
  ic.command = ISDN_STAT_PROT;
  ic.arg = DSS1_STAT_INVOKE_BRD;
  ic.parm.dss1_io.hl_id = id;
  ic.parm.dss1_io.ll_id = 0;
  ic.parm.dss1_io.proc = ident;
  ic.parm.dss1_io.timeout= 0;
  ic.parm.dss1_io.datalen = nlen;
  ic.parm.dss1_io.data = p;

  cs->iif.statcallb(&ic);
} /* l3dss1_nt_dummy_invoke */

static void
l3dss1_nt_parse_facility(struct PStack *st, struct l3_process *pc,
                      int cr, u_char * p)
{
	int qd_len = 0;
	unsigned char nlen = 0, ilen, cp_tag;
	int ident, id;
	ulong err_ret;

	if (pc) 
		st = pc->st; /* valid Stack */
	else
		if ((!st) || (cr >= 0)) return; /* neither pc nor st specified */

	p++;
	qd_len = *p++;
	if (qd_len == 0) {
		l3_debug(st, "qd_len == 0");
		return;
	}
	if ((*p & 0x1F) != 0x11) {	/* Service discriminator, supplementary service */
		l3_debug(st, "supplementary service != 0x11");
		return;
	}
	while (qd_len > 0 && !(*p & 0x80)) {	/* extension ? */
		p++;
		qd_len--;
	}
	if (qd_len < 2) {
		l3_debug(st, "qd_len < 2");
		return;
	}
	p++;
	qd_len--;
	if ((*p & 0xE0) != 0xA0) {	/* class and form */
		l3_debug(st, "class and form != 0xA0");
		return;
	}
       
        cp_tag = *p & 0x1F; /* remember tag value */

        p++;
	qd_len--;
	if (qd_len < 1) 
          { l3_debug(st, "qd_len < 1");
	    return;
	  }
	if (*p & 0x80) 
          { /* length format indefinite or limited */
	    nlen = *p++ & 0x7F; /* number of len bytes or indefinite */
            if ((qd_len-- < ((!nlen) ? 3 : (1 + nlen))) ||
                (nlen > 1))   
	     { l3_debug(st, "length format error or not implemented");
	       return;
             }
            if (nlen == 1)
	     { nlen = *p++; /* complete length */
               qd_len--;
             } 
            else
	     { qd_len -= 2; /* trailing null bytes */
               if ((*(p+qd_len)) || (*(p+qd_len+1)))
		{ l3_debug(st,"length format indefinite error");
                  return;
                }
               nlen = qd_len;
             }
	  }
        else
	  { nlen = *p++;
	    qd_len--;
          } 
	if (qd_len < nlen) 
          { l3_debug(st, "qd_len < nlen");
	    return;
	  }
	qd_len -= nlen;

	if (nlen < 2) 
          { l3_debug(st, "nlen < 2");
	    return;
	  }
        if (*p != 0x02) 
          {  /* invoke identifier tag */
	     l3_debug(st, "invoke identifier tag !=0x02");
	     return;
	  }
	p++;
	nlen--;
	if (*p & 0x80) 
          { /* length format */
	    l3_debug(st, "invoke id length format 2");
	    return;
	  }
	ilen = *p++;
	nlen--;
	if (ilen > nlen || ilen == 0) 
          { l3_debug(st, "ilen > nlen || ilen == 0");
	    return;
	  }
	nlen -= ilen;
	id = 0;
	while (ilen > 0) 
          { id = (id << 8) | (*p++ & 0xFF);	/* invoke identifier */
	    ilen--;
	  }

	switch (cp_tag) {	/* component tag */
		case 1:	/* invoke */
				if (nlen < 2) {
					l3_debug(st, "nlen < 2 22");
					return;
				}
				if (*p != 0x02) {	/* operation value */
					l3_debug(st, "operation value !=0x02");
					return;
				}
				p++;
				nlen--;
				ilen = *p++;
				nlen--;
				if (ilen > nlen || ilen == 0) {
					l3_debug(st, "ilen > nlen || ilen == 0 22");
					return;
				}
				nlen -= ilen;
				ident = 0;
				while (ilen > 0) {
					ident = (ident << 8) | (*p++ & 0xFF);
					ilen--;
				}

                                if (!pc) 
			         { l3dss1_nt_dummy_invoke(st, cr, id, ident, p, nlen);
                                   return;
                                 } 
#if HISAX_DE_AOC
			{

#define FOO1(s,a,b) \
	    while(nlen > 1) {		\
		    int ilen = p[1];	\
		    if(nlen < ilen+2) {	\
			    l3_debug(st, "FOO1  nlen < ilen+2"); \
			    return;		\
		    }			\
		    nlen -= ilen+2;		\
		    if((*p & 0xFF) == (a)) {	\
			    int nlen = ilen;	\
			    p += 2;		\
			    b;		\
		    } else {		\
			    p += ilen+2;	\
		    }			\
	    }

				switch (ident) {
					case 0x22:	/* during */
						FOO1("1A", 0x30, FOO1("1C", 0xA1, FOO1("1D", 0x30, FOO1("1E", 0x02, ( {
							       ident = 0;
							nlen = (nlen)?nlen:0; /* Make gcc happy */
							while (ilen > 0) {
														     ident = (ident << 8) | *p++;
								  ilen--;
									}
														     if (ident > pc->para.chargeinfo) {
														     pc->para.chargeinfo = ident;
														     st->l3.l3l4(st, CC_CHARGE | INDICATION, pc);
									}
														     if (st->l3.debug & L3_DEB_CHARGE) {
														     if (*(p + 2) == 0) {
														     l3_debug(st, "charging info during %d", pc->para.chargeinfo);
									}
								   else {
														     l3_debug(st, "charging info final %d", pc->para.chargeinfo);
									}
									}
									}
								    )))))
							break;
					case 0x24:	/* final */
						FOO1("2A", 0x30, FOO1("2B", 0x30, FOO1("2C", 0xA1, FOO1("2D", 0x30, FOO1("2E", 0x02, ( {
							       ident = 0;
							nlen = (nlen)?nlen:0; /* Make gcc happy */
							while (ilen > 0) {
																      ident = (ident << 8) | *p++;
								  ilen--;
									}
																      if (ident > pc->para.chargeinfo) {
																      pc->para.chargeinfo = ident;
																      st->l3.l3l4(st, CC_CHARGE | INDICATION, pc);
									}
																      if (st->l3.debug & L3_DEB_CHARGE) {
																      l3_debug(st, "charging info final %d", pc->para.chargeinfo);
									}
									}
								   ))))))
							break;
					default:
                                                       l3_debug(st, "invoke break invalid ident %02x",ident);
						break;
				}
#undef FOO1

			}
#else  /* not HISAX_DE_AOC */
                        l3_debug(st, "invoke break");
#endif /* not HISAX_DE_AOC */
			break;
		case 2:	/* return result */
			 /* if no process available handle separately */ 
                        if (!pc)
			 { if (cr == -1) 
                             l3dss1_nt_dummy_return_result(st, id, p, nlen);
                           return; 
                         }   
                        if ((pc->prot.dss1_nt.invoke_id) && (pc->prot.dss1_nt.invoke_id == id))
                          { /* Diversion successful */
                            nt_free_invoke_id(st,pc->prot.dss1_nt.invoke_id);
                            pc->prot.dss1_nt.remote_result = 0; /* success */     
                            pc->prot.dss1_nt.invoke_id = 0;
                            pc->redir_result = pc->prot.dss1_nt.remote_result; 
                            st->l3.l3l4(st, CC_REDIR | INDICATION, pc);                                  } /* Diversion successful */
                        else
                          l3_debug(st,"return error unknown identifier");
			break;
		case 3:	/* return error */
                            err_ret = 0;
	                    if (nlen < 2) 
                              { l3_debug(st, "return error nlen < 2");
	                        return;
	                      }
                            if (*p != 0x02) 
                              { /* result tag */
	                        l3_debug(st, "invoke error tag !=0x02");
	                        return;
	                      }
	                    p++;
	                    nlen--;
	                    if (*p > 4) 
                              { /* length format */
	                        l3_debug(st, "invoke return errlen > 4 ");
	                        return;
	                      }
	                    ilen = *p++;
	                    nlen--;
	                    if (ilen > nlen || ilen == 0) 
                              { l3_debug(st, "error return ilen > nlen || ilen == 0");
	                        return;
	                       }
	                    nlen -= ilen;
	                    while (ilen > 0) 
                             { err_ret = (err_ret << 8) | (*p++ & 0xFF);	/* error value */
	                       ilen--;
	                     }
			 /* if no process available handle separately */ 
                        if (!pc)
			 { if (cr == -1)
                             l3dss1_nt_dummy_error_return(st, id, err_ret);
                           return; 
                         }   
                        if ((pc->prot.dss1_nt.invoke_id) && (pc->prot.dss1_nt.invoke_id == id))
                          { /* Deflection error */
                            nt_free_invoke_id(st,pc->prot.dss1_nt.invoke_id);
                            pc->prot.dss1_nt.remote_result = err_ret; /* result */
                            pc->prot.dss1_nt.invoke_id = 0; 
                            pc->redir_result = pc->prot.dss1_nt.remote_result; 
                            st->l3.l3l4(st, CC_REDIR | INDICATION, pc);  
                          } /* Deflection error */
                        else
                          l3_debug(st,"return result unknown identifier");
			break;
		default:
			l3_debug(st, "facility default break tag=0x%02x",cp_tag);
			break;
	}
}

static void
l3dss1_nt_message(struct l3_process *pc, u_char mt)
{
	struct sk_buff *skb;
	u_char *p;

	if (!(skb = l3_alloc_skb(4)))
		return;
	p = skb_put(skb, 4);
	MsgHead(p, pc->callref, mt);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

static void
l3dss1_nt_message_cause(struct l3_process *pc, u_char mt, u_char cause)
{
	struct sk_buff *skb;
	u_char tmp[16];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, mt);
	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

static void
l3dss1_nt_status_send(struct l3_process *pc, u_char pr, void *arg)
{
	u_char tmp[16];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
	printk(KERN_DEBUG "DEBUG: status cref=%d state=%d cause=%d\n",pc->callref,pc->state,pc->para.cause);
	MsgHead(p, pc->callref, MT_STATUS);

	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = pc->para.cause | 0x80;

	*p++ = IE_CALL_STATE;
	*p++ = 0x1;
	*p++ = pc->state & 0x3f;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

static void
l3dss1_nt_msg_release_complete(struct l3_process *pc, u_char pr, void *arg)
{
	/* This routine is called if here was no SETUP made (checks in dss1up_nt and in
	 * l3dss1_nt_setup) and a RELEASE_COMPLETE have to be sent with an error code
	 * MT_STATUS_ENQUIRE in the NULL state is handled too
	 */
	u_char tmp[16];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;

	switch (pc->para.cause) {
		case 34:	/* no channel available */
		case 81:	/* invalid callreference */
		case 88:	/* incomp destination */
		case 96:	/* mandory IE missing */
		case 100:       /* invalid IE contents */
		case 101:	/* incompatible Callstate */
			MsgHead(p, pc->callref, MT_RELEASE_COMPLETE);
			*p++ = IE_CAUSE;
			*p++ = 0x2;
			*p++ = 0x81;
			*p++ = pc->para.cause | 0x80;
			break;
		default:
			printk(KERN_ERR "HiSax l3dss1_nt_msg_release_complete wrong cause %d\n",
				pc->para.cause);
			return;
	}
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	dss1_nt_release_l3_process(pc);
}

static void
l3dss1_nt_msg_resume_reject(struct l3_process *pc, u_char pr, void *arg)
{
	u_char tmp[16];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;

	MsgHead(p, pc->callref, MT_RESUME_REJECT);

	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = pc->para.cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	dss1_nt_release_l3_process(pc);
}
#if 0

static int ie_ALERTING[] = {IE_BEARER, IE_CHANNEL_ID | IE_MANDATORY_1,
		IE_FACILITY, IE_PROGRESS, IE_DISPLAY, IE_SIGNAL, IE_HLC,
		IE_USER_USER, -1};
static int ie_CALL_PROCEEDING[] = {IE_BEARER, IE_CHANNEL_ID | IE_MANDATORY_1,
		IE_FACILITY, IE_PROGRESS, IE_DISPLAY, IE_HLC, -1};
static int ie_CONNECT[] = {IE_BEARER, IE_CHANNEL_ID | IE_MANDATORY_1, 
		IE_FACILITY, IE_PROGRESS, IE_DISPLAY, IE_DATE, IE_SIGNAL,
		IE_CONNECT_PN, IE_CONNECT_SUB, IE_LLC, IE_HLC, IE_USER_USER, -1};
static int ie_CONNECT_ACKNOWLEDGE[] = {IE_CHANNEL_ID, IE_DISPLAY, IE_SIGNAL, -1};
#endif
static int ie_DISCONNECT[] = {IE_CAUSE | IE_MANDATORY, IE_FACILITY,
		IE_PROGRESS, IE_DISPLAY, IE_SIGNAL, IE_USER_USER, -1};
static int ie_INFORMATION[] = {IE_COMPLETE, IE_DISPLAY, IE_KEYPAD, IE_SIGNAL,
		IE_CALLED_PN, -1};
static int ie_NOTIFY[] = {IE_BEARER, IE_NOTIFY | IE_MANDATORY, IE_DISPLAY, IE_REDIR_DN, -1};
static int ie_PROGRESS[] = {IE_BEARER, IE_CAUSE, IE_FACILITY, IE_PROGRESS |
		IE_MANDATORY, IE_DISPLAY, IE_HLC, IE_USER_USER, -1};
static int ie_RELEASE[] = {IE_CAUSE | IE_MANDATORY_1, IE_FACILITY, IE_DISPLAY,
		IE_SIGNAL, IE_USER_USER, -1};
/* a RELEASE_COMPLETE with errors don't require special actions 
static int ie_RELEASE_COMPLETE[] = {IE_CAUSE | IE_MANDATORY_1, IE_DISPLAY, IE_SIGNAL, IE_USER_USER, -1};
*/
static int ie_SETUP[] = {IE_COMPLETE, IE_BEARER  | IE_MANDATORY,
		IE_CHANNEL_ID| IE_MANDATORY, IE_FACILITY, IE_PROGRESS,
		IE_NET_FAC, IE_DISPLAY, IE_KEYPAD, IE_SIGNAL, IE_CALLING_PN,
		IE_CALLING_SUB, IE_CALLED_PN, IE_CALLED_SUB, IE_REDIR_NR,
		IE_LLC, IE_HLC, IE_USER_USER, -1};
static int ie_RESUME[] = {IE_CALL_ID, -1};
static int ie_SUSPEND[] = {IE_CALL_ID, -1};
#if 0
static int ie_SETUP_ACKNOWLEDGE[] = {IE_CHANNEL_ID | IE_MANDATORY, IE_FACILITY,
		IE_PROGRESS, IE_DISPLAY, IE_SIGNAL, -1};
#endif
static int ie_STATUS[] = {IE_CAUSE | IE_MANDATORY, IE_CALL_STATE |
		IE_MANDATORY, IE_DISPLAY, -1};
#if 0
static int ie_STATUS_ENQUIRY[] = {IE_DISPLAY, -1};
/* not used 
 * static int ie_CONGESTION_CONTROL[] = {IE_CONGESTION | IE_MANDATORY,
 *		IE_CAUSE | IE_MANDATORY, IE_DISPLAY, -1};
 * static int ie_USER_INFORMATION[] = {IE_MORE_DATA, IE_USER_USER | IE_MANDATORY, -1};
 * static int ie_RESTART[] = {IE_CHANNEL_ID, IE_DISPLAY, IE_RESTART_IND |
 *		IE_MANDATORY, -1};
 */
static int ie_FACILITY[] = {IE_FACILITY | IE_MANDATORY, IE_DISPLAY, -1};
#endif
static int nt_comp_required[] = {1,2,3,5,6,7,9,10,11,14,15,25,-1};
static int nt_l3_valid_states[] = {0,1,2,3,4,6,7,8,9,10,11,12,15,17,19,22,25,-1};

struct ie_len {
	int ie;
	int len;
};

static
struct ie_len max_ie_len[] = {
	{IE_SEGMENT, 4},
	{IE_BEARER, 12},
	{IE_CAUSE, 32},
	{IE_CALL_ID, 10},
	{IE_CALL_STATE, 3},
	{IE_CHANNEL_ID,	34},
	{IE_FACILITY, 255},
	{IE_PROGRESS, 4},
	{IE_NET_FAC, 255},
	{IE_NOTIFY, 3},
	{IE_DISPLAY, 82},
	{IE_DATE, 8},
	{IE_KEYPAD, 34},
	{IE_SIGNAL, 3},
	{IE_INFORATE, 6},
	{IE_E2E_TDELAY, 11},
	{IE_TDELAY_SEL, 5},
	{IE_PACK_BINPARA, 3},
	{IE_PACK_WINSIZE, 4},
	{IE_PACK_SIZE, 4},
	{IE_CUG, 7},
	{IE_REV_CHARGE, 3},
	{IE_CALLING_PN, 24},
	{IE_CALLING_SUB, 23},
	{IE_CALLED_PN, 24},
	{IE_CALLED_SUB, 23},
	{IE_REDIR_NR, 255},
	{IE_REDIR_DN, 255},
	{IE_TRANS_SEL, 255},
	{IE_RESTART_IND, 3},
	{IE_LLC, 18},
	{IE_HLC, 5},
	{IE_USER_USER, 131},
	{-1,0},
};

static int
nt_getmax_ie_len(u_char ie) {
	int i = 0;
	while (max_ie_len[i].ie != -1) {
		if (max_ie_len[i].ie == ie)
			return(max_ie_len[i].len);
		i++;
	}
	return(255);
}

static int
nt_ie_in_set(struct l3_process *pc, u_char ie, int *checklist) {
	int ret = 1;

	while (*checklist != -1) {
		if ((*checklist & 0xff) == ie) {
			if (ie & 0x80)
				return(-ret);
			else
				return(ret);
		}
		ret++;
		checklist++;
	}
	return(0);
}

static int
nt_check_infoelements(struct l3_process *pc, struct sk_buff *skb, int *checklist)
{
	int *cl = checklist;
	u_char mt;
	u_char *p, ie;
	int l, newpos, oldpos;
	int err_seq = 0, err_len = 0, err_compr = 0, err_ureg = 0;
	u_char codeset = 0;
	u_char old_codeset = 0;
	u_char codelock = 1;
	
	p = skb->data;
	/* skip cr */
	p++;
	l = (*p++) & 0xf;
	p += l;
	mt = *p++;
	oldpos = 0;
	while ((p - skb->data) < skb->len) {
		if ((*p & 0xf0) == 0x90) { /* shift codeset */
			old_codeset = codeset;
			codeset = *p & 7;
			if (*p & 0x08)
				codelock = 0;
			else
				codelock = 1;
			if (pc->debug & L3_DEB_CHECK)
				l3_debug(pc->st, "check IE shift%scodeset %d->%d",
					codelock ? " locking ": " ", old_codeset, codeset);
			p++;
			continue;
		}
		if (!codeset) { /* only codeset 0 */
			if ((newpos = nt_ie_in_set(pc, *p, cl))) {
				if (newpos > 0) {
					if (newpos < oldpos)
						err_seq++;
					else
						oldpos = newpos;
				}
			} else {
				if (nt_ie_in_set(pc, *p, nt_comp_required))
					err_compr++;
				else
					err_ureg++;
			}
		}
		ie = *p++;
		if (ie & 0x80) {
			l = 1;
		} else {
			l = *p++;
			p += l;
			l += 2;
		}
		if (!codeset && (l > nt_getmax_ie_len(ie)))
			err_len++;
		if (!codelock) {
			if (pc->debug & L3_DEB_CHECK)
				l3_debug(pc->st, "check IE shift back codeset %d->%d",
					codeset, old_codeset);
			codeset = old_codeset;
			codelock = 1;
		}
	}
	if (err_compr | err_ureg | err_len | err_seq) {
		if (pc->debug & L3_DEB_CHECK)
			l3_debug(pc->st, "check IE MT(%x) %d/%d/%d/%d",
				mt, err_compr, err_ureg, err_len, err_seq);
		if (err_compr)
			return(ERR_IE_COMPREHENSION);
		if (err_ureg)
			return(ERR_IE_UNRECOGNIZED);
		if (err_len)
			return(ERR_IE_LENGTH);
		if (err_seq)
			return(ERR_IE_SEQUENCE);
	} 
	return(0);
}

/* verify if a message type exists and contain no IE error */
static int
l3dss1_nt_check_messagetype_validity(struct l3_process *pc, int mt, void *arg)
{
	switch (mt) {
		case MT_ALERTING:
		case MT_CALL_PROCEEDING:
		case MT_CONNECT:
		case MT_CONNECT_ACKNOWLEDGE:
		case MT_DISCONNECT:
		case MT_INFORMATION:
		case MT_FACILITY:
		case MT_NOTIFY:
		case MT_PROGRESS:
		case MT_RELEASE:
		case MT_RELEASE_COMPLETE:
		case MT_SETUP:
		case MT_SETUP_ACKNOWLEDGE:
		case MT_RESUME_ACKNOWLEDGE:
		case MT_RESUME_REJECT:
		case MT_SUSPEND_ACKNOWLEDGE:
		case MT_SUSPEND_REJECT:
		case MT_USER_INFORMATION:
		case MT_RESTART:
		case MT_RESTART_ACKNOWLEDGE:
		case MT_CONGESTION_CONTROL:
		case MT_STATUS:
		case MT_STATUS_ENQUIRY:
			if (pc->debug & L3_DEB_CHECK)
				l3_debug(pc->st, "l3dss1_nt_check_messagetype_validity mt(%x) OK", mt);
			break;
		case MT_RESUME: /* RESUME only in user->net */
		case MT_SUSPEND: /* SUSPEND only in user->net */
		default:
			if (pc->debug & (L3_DEB_CHECK | L3_DEB_WARN))
				l3_debug(pc->st, "l3dss1_nt_check_messagetype_validity mt(%x) fail", mt);
			pc->para.cause = 97;
			l3dss1_nt_status_send(pc, 0, NULL);
			return(1);
	}
	return(0);
}

static void
l3dss1_nt_std_ie_err(struct l3_process *pc, int ret) {

	if (pc->debug & L3_DEB_CHECK)
		l3_debug(pc->st, "nt_check_infoelements ret %d", ret);
	switch(ret) {
		case 0: 
			break;
		case ERR_IE_COMPREHENSION:
			pc->para.cause = 96;
			l3dss1_nt_status_send(pc, 0, NULL);
			break;
		case ERR_IE_UNRECOGNIZED:
			pc->para.cause = 99;
			l3dss1_nt_status_send(pc, 0, NULL);
			break;
		case ERR_IE_LENGTH:
			pc->para.cause = 100;
			l3dss1_nt_status_send(pc, 0, NULL);
			break;
		case ERR_IE_SEQUENCE:
		default:
			break;
	}
}

static int
l3dss1_nt_get_channel_id(struct l3_process *pc, struct sk_buff *skb) {
	u_char *p;

	p = skb->data;
	if ((p = findie(p, skb->len, IE_CHANNEL_ID, 0))) {
		p++;
		if (*p != 1) { /* len for BRI = 1 */
			if (pc->debug & L3_DEB_WARN)
				l3_debug(pc->st, "wrong chid len %d", *p);
			return (-2);
		}
		p++;
		if (*p & 0x60) { /* only base rate interface */
			if (pc->debug & L3_DEB_WARN)
				l3_debug(pc->st, "wrong chid %x", *p);
			return (-3);
		}
		return(*p & 0xb);
	} else
		return(-1);
}

static int
l3dss1_nt_get_cause(struct l3_process *pc, struct sk_buff *skb) {
	u_char l, i=0;
	u_char *p;

	p = skb->data;
	pc->para.cause = 31;
	pc->para.loc = 0;
	if ((p = findie(p, skb->len, IE_CAUSE, 0))) {
		p++;
		l = *p++;
		if (l>30)
			return(1);
		if (l) {
			pc->para.loc = *p++;
			l--;
		} else {
			return(2);
		}
		if (l && !(pc->para.loc & 0x80)) {
			l--;
			p++; /* skip recommendation */
		}
		if (l) {
			pc->para.cause = *p++;
			l--;
			if (!(pc->para.cause & 0x80))
				return(3);
		} else
			return(4);
		while (l && (i<6)) {
			pc->para.diag[i++] = *p++;
			l--;
		}
	} else
		return(-1);
	return(0);
}


#ifdef CONFIG_HISAX_SIGNALMODE
/* get_setupextensions
 * 
 * read the given skb data (skb) and parse,
 * the result is writen (to signal structure).
 */
static void get_setupextensions(struct l3_process *pc, struct sk_buff *skb)
{
	u_char *p;
	signal_msg *signal = &pc->para.signal;
//#warning remove the folowing debug lines
//printk(KERN_DEBUG "debug_nt 1\n");
	/* set default values */
	memset(signal, 0, sizeof(signal_msg));

//printk(KERN_DEBUG "debug_nt 2\n");
	/* called party number */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLED_PN, 0))) {
		iencpy(signal->called, p, sizeof(signal->called), 1);
		signal->called_type = ((p[2]&0x7f) >> 4) + 0x10;
		signal->called_plan = p[2] & 0xf;
	}
//printk(KERN_DEBUG "\n\ndebug_nt 3 called: '%s'\n\n\n", signal->called);

	/* called party subaddress */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLED_SUB, 0))) {
		iencpy(signal->called_sub, p, sizeof(signal->called_sub), 1);
		signal->called_subtype = ((p[2]&0x70) >> 4) + 0x10;
		signal->called_subodd = (p[2]&0x08) >> 3;
	}
//printk(KERN_DEBUG "debug_nt 4\n");

	/* called party keypad */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_KEYPAD, 0))) {
		iencpy(signal->keypad, p, sizeof(signal->keypad), 0);
	}
//printk(KERN_DEBUG "debug_nt 5\n");

	/* calling party number */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLING_PN, 0))) {
		signal->calling_type = ((p[2]&0x70) >> 4) + 0x10;
		signal->calling_plan = p[2] & 0xf;
		if (!(p[2] & 0x80)) {
			iencpy(signal->calling, p, sizeof(signal->calling), 2);
			signal->calling_screen = p[3] & 0x3;
			signal->calling_present = (p[3]&0x60) >> 5;
		} else {
			iencpy(signal->calling, p, sizeof(signal->calling), 1);
		}
	}
//printk(KERN_DEBUG "debug_nt 6\n");

	/* calling party subaddress */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLING_SUB, 0))) {
		iencpy(signal->calling_sub, p, sizeof(signal->calling_sub), 1);
		signal->calling_subtype = ((p[2]&0x70) >> 4) + 0x10;
		signal->calling_subodd = (p[2]&0x08) >> 3;
	}
//printk(KERN_DEBUG "debug_nt 7\n");

	/* connected party number */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CONNECT_PN, 0))) {
		signal->called_type = ((p[2]&0x7f) >> 4) + 0x10;
		signal->called_plan = p[2] & 0xf;
		if (!(p[2] & 0x80)) {
			iencpy(signal->called, p, sizeof(signal->called), 2);
			signal->called_screen = p[3] & 0x3;
			signal->called_present = (p[3]&0x60) >> 5;
		} else {
			iencpy(signal->called, p, sizeof(signal->called), 1);
		}
	}
//printk(KERN_DEBUG "debug_nt 8\n");

	/* connected party subaddress */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CONNECT_SUB, 0))) {
		iencpy(signal->called_sub, p, sizeof(signal->called_sub), 1);
		signal->called_subtype = ((p[2]&0x70) >> 4) + 0x10;
		signal->called_subodd = (p[2]&0x08) >> 3;
	}
//printk(KERN_DEBUG "debug_nt 9\n");

	/* redirecting party number */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_REDIR_NR, 0))) {
		signal->redirecting_type = ((p[2]&0x70) >> 4) + 0x10;
		signal->redirecting_plan = p[2] & 0xf;
		if (!(p[2] & 0x80)) {
			signal->redirecting_present = (p[3]&0x60) >> 5;
			if (!(p[3] & 0x80)) {
				iencpy(signal->redirecting, p, sizeof(signal->redirecting), 3);
				signal->redirecting_reason = p[4] & 0xf;
			} else {
				iencpy(signal->redirecting, p, sizeof(signal->redirecting), 2);
			}
		} else {
			iencpy(signal->redirecting, p, sizeof(signal->redirecting), 1);
		}
	}
//printk(KERN_DEBUG "debug_nt a\n");

	/* redirection party number */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_REDIR_DN, 0))) {
		signal->redirection_type = ((p[2]&0x70) >> 4) + 0x10;
		signal->redirection_plan = p[2] & 0xf;
		if (!(p[2] & 0x80)) {
			signal->redirection_present = (p[3]&0x60) >> 5;
			iencpy(signal->redirection, p, sizeof(signal->redirection), 2);
		} else {
			iencpy(signal->redirection, p, sizeof(signal->redirection), 1);
		}
	}
//printk(KERN_DEBUG "debug_nt b\n");

	/* notification indicator */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_NOTIFY, 0))) {
		signal->notify = p[2] & 0x7f;
	}
//printk(KERN_DEBUG "debug_nt c\n");

	/* bearer capacity */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_BEARER, 0))) {
		if (p[1] >= 2) {
			signal->bearer_coding = ((p[2]&0x60) >> 5);
			signal->bearer_capacity = 0x20 + (p[2] & 0x1f);
			signal->bearer_mode = ((p[3]&0x60) >> 5);
			signal->bearer_rate = (p[3] & 0x1f);
			if ((signal->bearer_rate&0x1f) == 0x18) {
				signal->bearer_ratemulti = (p[4] & 0x7f);
				if (p[1] >= 4) {
					signal->bearer_userinfo = (p[5] & 0x1f);
				}
			} else {
				if (p[1] >= 3) {
					signal->bearer_userinfo = (p[4] & 0x1f);
				}
			}
		}
	}
//printk(KERN_DEBUG "debug_nt d\n");

	/* sending complete */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_COMPLETE, 0))) {
		signal->sending_complete = 1;
	}
//printk(KERN_DEBUG "debug_nt e\n");

	/* cause */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CAUSE, 0))) {
		if (p[1] >= 2) {
			if (!(p[2] & 0x80)) {
				if (p[1] >= 3)
					signal->cause = (p[4] & 0x7f);
			} else
				signal->cause = (p[3] & 0x7f);
		}
	}

	/* call identity */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALL_ID, 0))) {
		signal->ident_length = p[1];
		if (signal->ident_length > 10)
			signal->ident_length = 10;
		memcpy(signal->ident, p+1, signal->ident_length);
	}

	/* prevent buffer overflow bugs */
	signal->called[sizeof(signal->called)-1] = '\0';
	signal->called_sub[sizeof(signal->called_sub)-1] = '\0';
	signal->keypad[sizeof(signal->keypad)-1] = '\0';
	signal->calling[sizeof(signal->calling)-1] = '\0';
	signal->calling_sub[sizeof(signal->calling_sub)-1] = '\0';
	signal->redirecting[sizeof(signal->redirecting)-1] = '\0';
	signal->redirection[sizeof(signal->redirection)-1] = '\0';
	signal->display[sizeof(signal->display)-1] = '\0';

//printk(KERN_DEBUG "debug_nt f\n");
}


/* put_setupextensions
 * 
 * read the given signal structure and appends information to the pointer p
 * the message defines which message to use. start and stop defines the range
 * of elements to append.
 */
static u_char *put_setupextensions(struct l3_process *pc, u_char *p, int message, int start, int stop)
{
	signal_msg *signal = &pc->para.signal;
	struct Channel *chan = pc->chan;

	if (!chan->signalmode)
		return(p);

	/* prevent buffer overflow bugs */
	signal->called[sizeof(signal->called)-1] = '\0';
	signal->called_sub[sizeof(signal->called_sub)-1] = '\0';
	signal->keypad[sizeof(signal->keypad)-1] = '\0';
	signal->calling[sizeof(signal->calling)-1] = '\0';
	signal->calling_sub[sizeof(signal->calling_sub)-1] = '\0';
	signal->redirecting[sizeof(signal->redirecting)-1] = '\0';
	signal->redirection[sizeof(signal->redirection)-1] = '\0';
	signal->display[sizeof(signal->display)-1] = '\0';

	/* NOTE: all elements must be given in ascending order of element identifier */

	if (message == MT_NOTIFY
	 || message == MT_SETUP
	 || message == MT_CALL_PROCEEDING
	 || message == MT_ALERTING
	 || message == MT_PROGRESS) {
		if ((signal->bearer_capacity & 0x20) && IE_BEARER>=start && IE_BEARER<=stop) {
			/* bearer capacity */
			*p++ = IE_BEARER;
			*p = 2;
			if ((signal->bearer_rate&0x1f) == 0x18) {
				(*p) ++;
			}
			if (signal->bearer_userinfo & 0x1f) {
				(*p) ++;
			}
			p++;
			*p++ = 0x80 + ((signal->bearer_coding&0x3) << 5) + (signal->bearer_capacity&0x1f);
			*p++ = 0x80 + ((signal->bearer_mode&0x3) << 5) + (signal->bearer_rate&0x1f);
			if ((signal->bearer_rate&0x1f) == 0x18) {
				*p++ = 0x80 + (signal->bearer_ratemulti & 0x7f);
			}
			if (signal->bearer_userinfo & 0x1f) {
				*p++ = 0xa0 + (signal->bearer_userinfo & 0x1f);
			}
		}
	}

	if (message == MT_DISCONNECT
	 || message == MT_PROGRESS
	 || message == MT_RELEASE
	 || message == MT_RELEASE_COMPLETE
	 || message == MT_RESUME_REJECT
	 || message == MT_SUSPEND_REJECT) {
		if (signal->cause && IE_CAUSE>=start && IE_CAUSE<=stop) {
			/* clear cause */
			*p++ = IE_CAUSE;
			*p++ = 0x2;
			*p++ = 0x81;
			*p++ = signal->cause | 0x80;
		}
	}

	if (message == MT_NOTIFY) {
		if (1 && IE_NOTIFY>=start && IE_NOTIFY<=stop) {
			/* notification indicator */
			*p++ = IE_NOTIFY;
			*p++ = 1;
			*p++ = 0x80 + (signal->notify & 0x7f);
		}
	}

	if (message == MT_ALERTING
	 || message == MT_CALL_PROCEEDING
	 || message == MT_CONNECT
	 || message == MT_CONNECT_ACKNOWLEDGE
	 || message == MT_DISCONNECT
	 || message == MT_NOTIFY
	 || message == MT_PROGRESS
	 || message == MT_SETUP
	 || message == MT_SETUP_ACKNOWLEDGE
	 || message == MT_INFORMATION
	 || message == MT_SUSPEND_ACKNOWLEDGE
	 || message == MT_SUSPEND_REJECT
	 || message == MT_RESUME_ACKNOWLEDGE
	 || message == MT_RESUME_REJECT
	 || message == MT_RELEASE) {
		/* send display */
		if (signal->display[0] && IE_DISPLAY>=start && IE_DISPLAY<=stop) {
			*p++ = IE_DISPLAY;
			*p++ = strlen(signal->display);
			strncpy(p, signal->display, strlen(signal->display));
			p += strlen(signal->display);
		}
	}

	if (message == MT_SETUP
	 || message == MT_INFORMATION) {
		/* called party keypad */
		if (signal->keypad[0] && IE_KEYPAD>=start && IE_KEYPAD<=stop) {
			*p++ = IE_KEYPAD;
			*p++ = strlen(signal->keypad);
			strncpy(p, signal->keypad, strlen(signal->keypad));
			p += strlen(signal->keypad);
		}
	}

	if (message == MT_ALERTING
	 || message == MT_CONNECT
	 || message == MT_CONNECT_ACKNOWLEDGE
	 || message == MT_DISCONNECT
	 || message == MT_INFORMATION
	 || message == MT_SETUP
	 || message == MT_SETUP_ACKNOWLEDGE
	 || message == MT_RELEASE) {
		/* send signal */
		if ((signal->signal >= 0x100) && IE_SIGNAL>=start && IE_SIGNAL<=stop) {
			*p++ = IE_SIGNAL;
			*p++ = 1;
			*p++ = signal->signal & 0xff;
		}
	}

	if (message == MT_CONNECT) {
		/* connected party number */
		if (signal->called_type && IE_CONNECT_PN>=start && IE_CONNECT_PN<=stop) {
			*p++ = IE_CONNECT_PN;
			*p++ = strlen(signal->called)+2;
			*p++ = 0x00 + ((signal->called_type&0x7)<<4) + (signal->called_plan&0xf);
			*p++ = 0x80 + ((signal->called_present&0x3)<<5) + (signal->called_screen&0x3);
			strncpy(p, signal->called, strlen(signal->called));
			p += strlen(signal->called);
		}

		/* connected party subaddress */
		if (signal->called_subtype && IE_CONNECT_SUB>=start && IE_CONNECT_SUB<=stop) {
			*p++ = IE_CONNECT_SUB;
			*p++ = strlen(signal->called_sub)+1;
			*p++ = 0x80 + ((signal->called_subtype&0x7)<<4) + ((signal->called_subodd&0x1)<<3);
			strncpy(p, signal->called_sub, strlen(signal->called_sub));
			p += strlen(signal->called_sub);
		}
	}

	if (message == MT_SETUP) {
		/* calling party number */
		if (signal->calling_type && IE_CALLING_PN>=start && IE_CALLING_PN<=stop) {
			*p++ = IE_CALLING_PN;
			*p++ = strlen(signal->calling)+2;
			*p++ = 0x00 + ((signal->calling_type&0x7)<<4) + (signal->calling_plan&0xf);
			*p++ = 0x80 + ((signal->calling_present&0x3)<<5) + (signal->calling_screen&0x3);
			strncpy(p, signal->calling, strlen(signal->calling));
			p += strlen(signal->calling);
		}

		/* calling party subaddress */
		if (signal->calling_subtype && IE_CALLING_SUB>=start && IE_CALLING_SUB<=stop) {
			*p++ = IE_CALLING_SUB;
			*p++ = strlen(signal->calling_sub)+1;
			*p++ = 0x80 + ((signal->calling_subtype&0x7)<<4) + ((signal->calling_subodd&0x1)<<3);
			strncpy(p, signal->calling_sub, strlen(signal->calling_sub));
			p += strlen(signal->calling_sub);
		}
	}

	if (message == MT_SETUP
	 || message == MT_INFORMATION) {
		/* called party number */
		if (signal->called_type && IE_CALLED_PN>=start && IE_CALLED_PN<=stop) {
			*p++ = IE_CALLED_PN;
			*p++ = strlen(signal->called)+1;
			*p++ = 0x80 + ((signal->called_type&0x7)<<4) + (signal->called_plan&0xf);
			strncpy(p, signal->called, strlen(signal->called));
			p += strlen(signal->called);
		}

		/* called party subaddress */
		if (signal->called_subtype && IE_CALLED_SUB>=start && IE_CALLED_SUB<=stop) {
			*p++ = IE_CALLED_SUB;
			*p++ = strlen(signal->called_sub)+1;
			*p++ = 0x80 + ((signal->called_subtype&0x7)<<4) + ((signal->called_subodd&0x1)<<3);
			strncpy(p, signal->called_sub, strlen(signal->called_sub));
			p += strlen(signal->called_sub);
		}
	}

	if (message == MT_SETUP) {
		/* redirecting party number */
		if (signal->redirecting_type && IE_REDIR_NR>=start && IE_REDIR_NR<=stop) {
			*p++ = IE_REDIR_NR;
			*p++ = strlen(signal->redirecting)+3;
			*p++ = 0x00 + ((signal->redirecting_type&0x7)<<4) + (signal->redirecting_plan&0xf);
			*p++ = 0x00 + ((signal->redirecting_present&0x3)<<5);
			*p++ = 0x80 + (signal->redirecting_reason & 0xf);
			strncpy(p, signal->redirecting, strlen(signal->redirecting));
			p += strlen(signal->redirecting);
		}
	}

	if (message == MT_NOTIFY) {
		/* redirection party number */
		if (signal->redirection[0] && IE_REDIR_DN>=start && IE_REDIR_DN<=stop) {
			*p++ = IE_REDIR_DN;
			*p++ = strlen(signal->redirection)+2;
			*p++ = 0x00 + ((signal->redirection_type&0x7)<<4) + (signal->redirection_plan&0xf);
			*p++ = 0x80 + ((signal->redirection_present&0x3)<<5);
			strncpy(p, signal->redirection, strlen(signal->redirection));
			p += strlen(signal->redirection);
		}
	}

	if (message == MT_SETUP
	 || message == MT_INFORMATION) {
		if (signal->sending_complete && IE_COMPLETE>=start && IE_COMPLETE<=stop) {
			/* sending complete */
			*p++ = IE_COMPLETE;
		}
	}

	/* return new pointer */
	return(p);
}


#endif
/* request free bchannel and return 0 if not available
the bchannel is stored in pc->para.bchannel
*/
inline int
l3_get_bchannel(struct l3_process *pc, u_char pr)
{
	struct IsdnCardState *cs = pc->st->l1.hardware;
	int id;

	id=pc->para.bchannel;
	if (id >= 0) { /* channel information given */
		if (id & 0x8) { /* exclusive */
			id = bchannel_nt_alloc(cs,id & 0x3);
		} else { /* prefered */
			id = bchannel_nt_alloc(cs,id & 0x3);
			if (id <= 0)
				id = bchannel_nt_alloc(cs,-1);
		}
		if (id <= 0) { /* error in getting bchannel */
			if (pc->debug & L3_DEB_WARN)
				l3_debug(pc->st, "setup with b-channel %d is in use or not specified",id);
			pc->para.cause = 34;
			l3dss1_nt_msg_release_complete(pc, pr, NULL);
			return(0);
		} else {
			if (pc->debug & L3_DEB_STATE)
				l3_debug(pc->st, "assigning d-channel [1..2] = %d",id);
		}
		pc->para.bchannel = id;
	} else {
		/* see which channel is free and reject call if not available */
		id = bchannel_nt_alloc(cs,-1);
		if (id <= 0) {
			if (pc->debug & L3_DEB_WARN)
				l3_debug(pc->st, "no b-channel available for setup");
			pc->para.cause = 34;
			l3dss1_nt_msg_release_complete(pc, pr, NULL);
			return(0);
		}
		pc->para.bchannel = id;
	}
	return(1);
}

#if 0
static void
l3dss1_nt_msg_with_uus(struct l3_process *pc, u_char cmd)
{
	struct sk_buff *skb;
	u_char tmp[16+40];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, cmd);

        if (pc->prot.dss1_nt.uus1_data[0])
	 { *p++ = IE_USER_USER; /* UUS info element */
           *p++ = strlen(pc->prot.dss1_nt.uus1_data) + 1;
           *p++ = 0x04; /* IA5 chars */
           strcpy(p,pc->prot.dss1_nt.uus1_data);
           p += strlen(pc->prot.dss1_nt.uus1_data);
           pc->prot.dss1_nt.uus1_data[0] = '\0';   
         } 

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
} /* l3dss1_nt_msg_with_uus */
#endif

/* is calles when some state get a release request or
state 12 gets a T305/T306 timeout */
static void
l3dss1_nt_release_req(struct l3_process *pc, u_char pr, void *arg)
{
	/* stop all timers */
	StopAllL3Timer(pc);

	/* change to state 19 */
	newl3state(pc, 19);

	/* send release message to user */
//	if (!pc->prot.dss1_nt.uus1_data[0]) 
	l3dss1_nt_message(pc, MT_RELEASE);
//	else
//		l3dss1_nt_msg_with_uus(pc, MT_RELEASE);

	/* start T308 */
	L3AddTimer(&pc->timer, T308, CC_T308_1);
}

/* process releases */

static void
l3dss1_nt_release_cmpl(struct l3_process *pc, u_char pr, void *arg)
{
	/* stop timer T308 */
	StopAllL3Timer(pc);

	/* disconnect bchannel */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_NONE);

	/* send confirm to call control */
	pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);

	/* change to state NULL */
	newl3state(pc, 0);

	/* release cr + bchannel */
	dss1_nt_release_l3_process(pc);
}

#if EXT_BEARER_CAPS

static u_char *
EncodeASyncParams(u_char * p, u_char si2)
{				// 7c 06 88  90 21 42 00 bb

	p[0] = 0;
	p[1] = 0x40;		// Intermediate rate: 16 kbit/s jj 2000.02.19
	p[2] = 0x80;
	if (si2 & 32)		// 7 data bits

		p[2] += 16;
	else			// 8 data bits

		p[2] += 24;

	if (si2 & 16)		// 2 stop bits

		p[2] += 96;
	else			// 1 stop bit

		p[2] += 32;

	if (si2 & 8)		// even parity

		p[2] += 2;
	else			// no parity

		p[2] += 3;

	switch (si2 & 0x07) {
		case 0:
			p[0] = 66;	// 1200 bit/s

			break;
		case 1:
			p[0] = 88;	// 1200/75 bit/s

			break;
		case 2:
			p[0] = 87;	// 75/1200 bit/s

			break;
		case 3:
			p[0] = 67;	// 2400 bit/s

			break;
		case 4:
			p[0] = 69;	// 4800 bit/s

			break;
		case 5:
			p[0] = 72;	// 9600 bit/s

			break;
		case 6:
			p[0] = 73;	// 14400 bit/s

			break;
		case 7:
			p[0] = 75;	// 19200 bit/s

			break;
	}
	return p + 3;
}

static  u_char
EncodeSyncParams(u_char si2, u_char ai)
{

	switch (si2) {
		case 0:
			return ai + 2;	// 1200 bit/s

		case 1:
			return ai + 24;		// 1200/75 bit/s

		case 2:
			return ai + 23;		// 75/1200 bit/s

		case 3:
			return ai + 3;	// 2400 bit/s

		case 4:
			return ai + 5;	// 4800 bit/s

		case 5:
			return ai + 8;	// 9600 bit/s

		case 6:
			return ai + 9;	// 14400 bit/s

		case 7:
			return ai + 11;		// 19200 bit/s

		case 8:
			return ai + 14;		// 48000 bit/s

		case 9:
			return ai + 15;		// 56000 bit/s

		case 15:
			return ai + 40;		// negotiate bit/s

		default:
			break;
	}
	return ai;
}

static u_char
nt_DecodeASyncParams(u_char si2, u_char * p)
{
	u_char info;

	switch (p[5]) {
		case 66:	// 1200 bit/s

			break;	// si2 don't change

		case 88:	// 1200/75 bit/s

			si2 += 1;
			break;
		case 87:	// 75/1200 bit/s

			si2 += 2;
			break;
		case 67:	// 2400 bit/s

			si2 += 3;
			break;
		case 69:	// 4800 bit/s

			si2 += 4;
			break;
		case 72:	// 9600 bit/s

			si2 += 5;
			break;
		case 73:	// 14400 bit/s

			si2 += 6;
			break;
		case 75:	// 19200 bit/s

			si2 += 7;
			break;
	}

	info = p[7] & 0x7f;
	if ((info & 16) && (!(info & 8)))	// 7 data bits

		si2 += 32;	// else 8 data bits

	if ((info & 96) == 96)	// 2 stop bits

		si2 += 16;	// else 1 stop bit

	if ((info & 2) && (!(info & 1)))	// even parity

		si2 += 8;	// else no parity

	return si2;
}


static u_char
nt_DecodeSyncParams(u_char si2, u_char info)
{
	info &= 0x7f;
	switch (info) {
		case 40:	// bit/s negotiation failed  ai := 165 not 175!

			return si2 + 15;
		case 15:	// 56000 bit/s failed, ai := 0 not 169 !

			return si2 + 9;
		case 14:	// 48000 bit/s

			return si2 + 8;
		case 11:	// 19200 bit/s

			return si2 + 7;
		case 9:	// 14400 bit/s

			return si2 + 6;
		case 8:	// 9600  bit/s

			return si2 + 5;
		case 5:	// 4800  bit/s

			return si2 + 4;
		case 3:	// 2400  bit/s

			return si2 + 3;
		case 23:	// 75/1200 bit/s

			return si2 + 2;
		case 24:	// 1200/75 bit/s

			return si2 + 1;
		default:	// 1200 bit/s

			return si2;
	}
}

#endif
static u_char
nt_DecodeSI2(struct sk_buff *skb)
{
	u_char *p;		//, *pend=skb->data + skb->len;

	if ((p = findie(skb->data, skb->len, 0x7c, 0))) {
		switch (p[4] & 0x0f) {
			case 0x01:
				if (p[1] == 0x04)	// sync. Bitratenadaption

					return nt_DecodeSyncParams(160, p[5]);	// V.110/X.30

				else if (p[1] == 0x06)	// async. Bitratenadaption

					return nt_DecodeASyncParams(192, p);	// V.110/X.30

				break;
			case 0x08:	// if (p[5] == 0x02) // sync. Bitratenadaption
				if (p[1] > 3) 
					return nt_DecodeSyncParams(176, p[5]);	// V.120
				break;
		}
	}
	return 0;
}

static void l3dss1_nt_setup_req_cont(struct l3_process *pc, u_char pr, void *arg);
static void
l3dss1_nt_setup_req(struct l3_process *pc, u_char pr,
		 void *arg)
{
	/* make this the first time of setup message */
	pc->N303 = 1; /* first timer start */

	/* alloc free bchannel */
	pc->para.bchannel = -1; /* any free bchannel */
	pc->para.bchannel_inuse = 0;
	if (!l3_get_bchannel(pc,pr)) {
		/* setup error */
#ifdef CONFIG_HISAX_SIGNALMODE
		memset(&pc->para.signal, 0, sizeof(signal_msg));
		pc->para.signal.cause = 34; /* no channel */
#endif
		pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);

		/* change to state NULL */
		newl3state(pc, 0);

		/* release cr + bchannel */
		dss1_nt_release_l3_process(pc);
		return;
	}
	pc->para.bchannel_inuse = 1;

	/* continue with setup */
	l3dss1_nt_setup_req_cont(pc, pr, arg);
}
/* NOTE: this is the second part of l3dss1_nt_setup_req */
static void
l3dss1_nt_setup_req_cont(struct l3_process *pc, u_char pr,
		 void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	u_char channel = 0, type = 4;

        u_char send_keypad;
	u_char screen = 0x80;
	u_char *teln;
	u_char *msn;
	u_char *sub;
	u_char *sp;
	int l;

	/* create message */
	channel = pc->para.bchannel + 0x88;
	*p++ = (CTRL_SAPI << 2) | 0x2;
	*p++ = (GROUP_TEI << 1) | 0x1;
	*p++ = UI;
	MsgHead(p, pc->callref, MT_SETUP);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (!pc->chan->signalmode) { /**/
#endif
	teln = pc->para.setup.phone;
#ifndef CONFIG_HISAX_NO_KEYPAD
        send_keypad = (strchr(teln,'*') || strchr(teln,'#')) ? 1 : 0; 
#else
	send_keypad = 0;
#endif
	if (42)
/* NOTE: MUST COMPLETE since we have not implemented the ptp in nt mode */
		*p++ = 0xa1;		/* complete indicator */
	/*
	 * Set Bearer Capability, Map info from 1TR6-convention to EDSS1
	 */
	switch (pc->para.setup.si1) {
	case 1:	                  /* Telephony                                */
		*p++ = IE_BEARER;
		*p++ = 0x3;	  /* Length                                   */
		*p++ = 0x90;	  /* Coding Std. CCITT, 3.1 kHz audio         */
		*p++ = 0x90;	  /* Circuit-Mode 64kbps                      */
		*p++ = 0xa3;	  /* A-Law Audio                              */
		break;
	case 5:	                  /* Datatransmission 64k, BTX                */
	case 7:	                  /* Datatransmission 64k                     */
	default:
		*p++ = IE_BEARER;
		*p++ = 0x2;	  /* Length                                   */
		*p++ = 0x88;	  /* Coding Std. CCITT, unrestr. dig. Inform. */
		*p++ = 0x90;	  /* Circuit-Mode 64kbps                      */
		break;
	}

	if (send_keypad) {
		*p++ = IE_KEYPAD;
		*p++ = strlen(teln);
		while (*teln)
			*p++ = (*teln++) & 0x7F;
	}
	  
	/*
	 * What about info2? Mapping to High-Layer-Compatibility?
	 */
	if ((*teln) && (!send_keypad)) {
		// parse number for special things 
		if (!isdigit(*teln)) {
			switch (0x5f & *teln) {
				case 'C':
				case 'P':
					teln++;
					break;
				case 'R':
					screen = 0xA0;
					break;
				case 'D':
					screen = 0x80;
					break;
				
			        default:
					if (pc->debug & L3_DEB_WARN)
						l3_debug(pc->st, "Wrong TELN Code");
					break;
			}
			teln++;
		}
	}

	if (channel) {
		*p++ = IE_CHANNEL_ID;
		*p++ = 1;
		*p++ = channel;
	}
	msn = pc->para.setup.eazmsn;
	if (*msn) {
		if (!isdigit(*msn)) {
			switch (0x5f & *msn) {
				case 'R':
					type = 7;
					break;
				case 'A':
					type = 6;
					break;
				case 'S':
					type = 4;
					break;
				case 'X':
					type = 3;
					break;
				case 'N':
					type = 2;
					break;
				case 'I':
					type = 1;
					break;
				case 'U':
					type = 0;
					break;
			        default:
					if (pc->debug & L3_DEB_WARN)
						l3_debug(pc->st, "Wrong MSN Code");
					break;
			}
			msn++;
		}
	}
	sub = NULL;
	sp = msn;
	while (*sp) {
		if ('.' == *sp) {
			sub = sp;
			*sp = 0;
		} else
			sp++;
	}
	if (*msn) {
		*p++ = IE_CALLING_PN;
		*p++ = strlen(msn) + (screen ? 2 : 1);
		/* Classify as AnyPref. */
		if (screen) {
			*p++ = 0x01 + (type<<4); /* Ext = '0'B, Type = '000'B, Plan = '0001'B. */
			*p++ = screen;
		} else
			*p++ = 0x81 + (type<<4); /* Ext = '1'B, Type = '000'B, Plan = '0001'B. */
		while (*msn)
			*p++ = *msn++ & 0x7f;
	}
	if (sub) {
		*sub++ = '.';
		*p++ = IE_CALLING_SUB;
		*p++ = strlen(sub) + 2;
		*p++ = 0x80;	/* NSAP coded */
		*p++ = 0x50;	/* local IDI format */
		while (*sub)
			*p++ = *sub++ & 0x7f;
	}
	sub = NULL;
	sp = teln;
	while (*sp) {
		if ('.' == *sp) {
			sub = sp;
			*sp = 0;
		} else
			sp++;
	}
	
        if (!send_keypad) {      
		*p++ = IE_CALLED_PN;
		*p++ = strlen(teln) + 1;
		/* Classify as AnyPref. */
		*p++ = 0xC1;		/* Ext = '1'B, Type = '100'B, Plan = '0001'B. */
		while (*teln)
			*p++ = *teln++ & 0x7f;
		
		if (sub) {
			*sub++ = '.';
			*p++ = IE_CALLED_SUB;
			*p++ = strlen(sub) + 2;
			*p++ = 0x80;	/* NSAP coded */
			*p++ = 0x50;	/* local IDI format */
			while (*sub)
				*p++ = *sub++ & 0x7f;
		}
        }
#if EXT_BEARER_CAPS
	if ((pc->para.setup.si2 >= 160) && (pc->para.setup.si2 <= 175)) {	// sync. Bitratenadaption, V.110/X.30

		*p++ = IE_LLC;
		*p++ = 0x04;
		*p++ = 0x88;
		*p++ = 0x90;
		*p++ = 0x21;
		*p++ = EncodeSyncParams(pc->para.setup.si2 - 160, 0x80);
	} else if ((pc->para.setup.si2 >= 176) && (pc->para.setup.si2 <= 191)) {	// sync. Bitratenadaption, V.120

		*p++ = IE_LLC;
		*p++ = 0x05;
		*p++ = 0x88;
		*p++ = 0x90;
		*p++ = 0x28;
		*p++ = EncodeSyncParams(pc->para.setup.si2 - 176, 0);
		*p++ = 0x82;
	} else if (pc->para.setup.si2 >= 192) {		// async. Bitratenadaption, V.110/X.30

		*p++ = IE_LLC;
		*p++ = 0x06;
		*p++ = 0x88;
		*p++ = 0x90;
		*p++ = 0x21;
		p = EncodeASyncParams(p, pc->para.setup.si2 - 192);
#ifndef CONFIG_HISAX_NO_LLC
	} else {
	  switch (pc->para.setup.si1) {
		case 1:	                /* Telephony                                */
			*p++ = IE_LLC;
			*p++ = 0x3;	/* Length                                   */
			*p++ = 0x90;	/* Coding Std. CCITT, 3.1 kHz audio         */
			*p++ = 0x90;	/* Circuit-Mode 64kbps                      */
			*p++ = 0xa3;	/* A-Law Audio                              */
			break;
		case 5:	                /* Datatransmission 64k, BTX                */
		case 7:	                /* Datatransmission 64k                     */
		default:
			*p++ = IE_LLC;
			*p++ = 0x2;	/* Length                                   */
			*p++ = 0x88;	/* Coding Std. CCITT, unrestr. dig. Inform. */
			*p++ = 0x90;	/* Circuit-Mode 64kbps                      */
			break;
	  }
#endif
	}
#endif
#ifdef CONFIG_HISAX_SIGNALMODE
	} /**/
	else {
		if (42) /* we assume that it is point-to-multipoint so we need to sending-complete. ptp is not supported by this implementation yet. */
			pc->para.signal.sending_complete = 1;
		/* we must add the channel because it cannot be included in
		   the setup extension */
		p = put_setupextensions(pc, p, MT_SETUP, 0, IE_CHANNEL_ID-1);
		if (channel) { /* there is always a channel */
			*p++ = IE_CHANNEL_ID;
			*p++ = 1;
			*p++ = channel;
		}
		p = put_setupextensions(pc, p, MT_SETUP, IE_CHANNEL_ID+1, 255);
	}
#endif

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);

	/* setup ack timers */
	L3DelTimer(&pc->timer);
	L3AddTimer(&pc->timer, T303, CC_T303);

	/* go into call present state */
	newl3state(pc, 6);

	/* we use the following hack: directly send data 
	   this is because we do not want to build any layer 2 connection
	   since broadcasting doesn't need a layer 2 connection of a TE */
	pc->st->l2.l2l1(pc->st, PH_DATA | REQUEST, skb);

}

/* call is proceeding after broadcasting setup
 */
static void
l3dss1_nt_call_proc(struct l3_process *pc, u_char pr, void *arg)
{
  	struct IsdnCardState *cs = pc->st->l1.hardware;

	/* be sure that link is up */
	l3_msg(pc->st, DL_ESTABLISH | REQUEST, NULL);

	/* stop T303 */
	L3DelTimer(&pc->timer);

	/* start T310 */
	if (cs->nt.outproc_tout)
		L3AddTimer(&pc->timer, cs->nt.outproc_tout, CC_Toutproc);

	/* outgoing call proceeding */
	newl3state(pc, 9);

	/* indicate proceeding */
	pc->st->l3.l3l4(pc->st, CC_PROCEEDING | INDICATION, pc);
}

/* setup ack after broadcasting setup
 */
static void
l3dss1_nt_setup_ack(struct l3_process *pc, u_char pr, void *arg)
{
	/* be sure that link is up */
	l3_msg(pc->st, DL_ESTABLISH | REQUEST, NULL);

	/* stop T303 */
	L3DelTimer(&pc->timer);

	/* start T304 */
	L3AddTimer(&pc->timer, T304, CC_T304);

	/* change to overlap recv */
	newl3state(pc, 25);

	/* indicate more information */
	pc->st->l3.l3l4(pc->st, CC_MORE_INFO | INDICATION, pc);
}

static void
l3dss1_nt_connect(struct l3_process *pc, u_char pr, void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb = arg;

	/* get COLP */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CONNECT_PN, 0))) {
		pc->para.setup.plan = p[2];
		if (p[2] & 0x80) {
			iecpy(pc->para.setup.phone, p, 1);
			pc->para.setup.screen = 0;
		} else {
			iecpy(pc->para.setup.phone, p, 2);
			pc->para.setup.screen = p[3];
		}
	} else {
		pc->para.setup.phone[0] = 0;
		pc->para.setup.plan = 0;
		pc->para.setup.screen = 0;
	}
	/* get COLPSUB */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CONNECT_SUB, 0))) {
		if ((p[1] >= 2) && (p[2] == 0x80) && (p[3] == 0x50)) {
			tmp[0] = '.';
			iecpy(&tmp[1], p, 2);
			if (strlen(pc->para.setup.phone)+strlen(tmp) < sizeof(pc->para.setup.phone))
				strcat(pc->para.setup.phone, tmp);
		} else if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "wrong connect subaddress");
	}

	/* stop timer */
	L3DelTimer(&pc->timer);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode) { 
		/* in signal mode we do not acknowledge the connect because
		   this is done in user space to send additional information
		   with the acknowledge. */
		newl3state(pc, 8);
	} else { /**/

#endif
	/* go directly to connected state */
	newl3state(pc, 10);

	/* reply directly with connect acknowledge */
	MsgHead(p, pc->callref, MT_CONNECT_ACKNOWLEDGE);

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

#ifdef CONFIG_HISAX_SIGNALMODE
	} /**/
#endif
	/* indicate setup confirm */
	pc->st->l3.l3l4(pc->st, CC_SETUP | CONFIRM, pc);
}

/* alerting at state 6 or 9 */
static void
l3dss1_nt_alerting(struct l3_process *pc, u_char pr, void *arg)
{
  	struct IsdnCardState *cs = pc->st->l1.hardware;

	/* be sure that link is up */
	l3_msg(pc->st, DL_ESTABLISH | REQUEST, NULL);

	/* stop timer T303 or T310 */
	L3DelTimer(&pc->timer);

	/* add timer T301 */
	if (cs->nt.outalert_tout)
		L3AddTimer(&pc->timer, cs->nt.outalert_tout, CC_Toutalert);

	/* change to call received */
	newl3state(pc, 7);

	/* indicate alerting */
	pc->st->l3.l3l4(pc->st, CC_ALERTING | INDICATION, pc);
}

static void
l3dss1_nt_setup(struct l3_process *pc, u_char pr, void *arg)
{
	u_char *p;
	char tmp[80];
	struct sk_buff *skb = arg;
	int err = 0;

	/*
	 * Bearer Capabilities
	 */
	p = skb->data;
	/* only the first occurence 'll be detected ! */
	if ((p = findie(p, skb->len, 0x04, 0))) {
		if ((p[1] < 2) || (p[1] > 11))
			err = 1;
		else {
			pc->para.setup.si2 = 0;
			switch (p[2] & 0x7f) {
				case 0x00: /* Speech */
				case 0x10: /* 3.1 Khz audio */
					pc->para.setup.si1 = 1;
					break;
				case 0x08: /* Unrestricted digital information */
					pc->para.setup.si1 = 7;
/* JIM, 05.11.97 I wanna set service indicator 2 */
#if EXT_BEARER_CAPS
					pc->para.setup.si2 = nt_DecodeSI2(skb);
#endif
					break;
				case 0x09: /* Restricted digital information */
					pc->para.setup.si1 = 2;
					break;
				case 0x11:
					/* Unrestr. digital information  with 
					 * tones/announcements ( or 7 kHz audio
					 */
					pc->para.setup.si1 = 3;
					break;
				case 0x18: /* Video */
					pc->para.setup.si1 = 4;
					break;
				default:
					err = 2;
					break;
			}
			switch (p[3] & 0x7f) {
				case 0x40: /* packed mode */
					pc->para.setup.si1 = 8;
					break;
				case 0x10: /* 64 kbit */
				case 0x11: /* 2*64 kbit */
				case 0x13: /* 384 kbit */
				case 0x15: /* 1536 kbit */
				case 0x17: /* 1920 kbit */
					pc->para.moderate = p[3] & 0x7f;
					break;
				default:
					err = 3;
					break;
			}
		}
		if (pc->debug & L3_DEB_SI)
			l3_debug(pc->st, "SI=%d, AI=%d",
				pc->para.setup.si1, pc->para.setup.si2);
		if (err) {
			if (pc->debug & L3_DEB_WARN)
				l3_debug(pc->st, "setup with wrong bearer(l=%d:%x,%x)",
					p[1], p[2], p[3]);
			pc->para.cause = 100;
			l3dss1_nt_msg_release_complete(pc, pr, NULL);
			return;
		}
	} else {
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "setup without bearer capabilities");
		/* ETS 300-104 1.3.3 */
		pc->para.cause = 96;
		l3dss1_nt_msg_release_complete(pc, pr, NULL);
		return;
	}

	pc->para.bchannel = l3dss1_nt_get_channel_id(pc, skb);
	pc->para.bchannel_inuse = 0;
	if (!l3_get_bchannel(pc,pr)) {
		pc->para.cause = 34; /* no channel */
		l3dss1_nt_msg_release_complete(pc, pr, NULL);
		return;
	}
	pc->para.bchannel_inuse = 1;

	/* Now we are on none mandatory IEs */
	err = nt_check_infoelements(pc, skb, ie_SETUP);
	if (ERR_IE_COMPREHENSION == err) {
		pc->para.cause = 96;
		l3dss1_nt_msg_release_complete(pc, pr, NULL);
		return;
	}

	/* called party */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLED_PN, 0))) {
		iecpy(pc->para.setup.eazmsn, p, 1);
	} else {
		pc->para.setup.eazmsn[0] = 0;
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLED_SUB, 0))) {
		/* Called party subaddress */
		if ((p[1] >= 2) && (p[2] == 0x80) && (p[3] == 0x50)) {
			tmp[0] = '.';
			iecpy(&tmp[1], p, 2);
			if (strlen(pc->para.setup.eazmsn)+strlen(tmp) < sizeof(pc->para.setup.eazmsn))
				strcat(pc->para.setup.eazmsn, tmp);
		} else if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "wrong called subaddress");
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_KEYPAD, 0))) {
		/* keypad */
		iecpy(tmp, p, 0);
		if (strlen(pc->para.setup.eazmsn)+strlen(tmp) < sizeof(pc->para.setup.eazmsn))
			strcat(pc->para.setup.eazmsn, tmp);
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLING_PN, 0))) {
		pc->para.setup.plan = p[2];
		if (p[2] & 0x80) {
			iecpy(pc->para.setup.phone, p, 1);
			pc->para.setup.screen = 0;
		} else {
			iecpy(pc->para.setup.phone, p, 2);
			pc->para.setup.screen = p[3];
		}
	} else {
		pc->para.setup.phone[0] = 0;
		pc->para.setup.plan = 0;
		pc->para.setup.screen = 0;
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLING_SUB, 0))) {
		/* Calling party subaddress */
		if ((p[1] >= 2) && (p[2] == 0x80) && (p[3] == 0x50)) {
			tmp[0] = '.';
			iecpy(&tmp[1], p, 2);
			strcat(pc->para.setup.phone, tmp);
		} else if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "wrong calling subaddress");
	}

	newl3state(pc, 1); /* call initiated */
	if (err) /* STATUS for none mandatory IE errors after actions are taken */
		l3dss1_nt_std_ie_err(pc, err);
	/* tell call control about setup */
	pc->st->l3.l3l4(pc->st, CC_SETUP | INDICATION, pc);
}

static void
l3dss1_nt_resume(struct l3_process *pc, u_char pr, void *arg)
{
	u_char *p;
	struct sk_buff *skb = arg;
	int err = 0;

	/* only signalmode supports resume */
#ifdef CONFIG_HISAX_SIGNALMODE
	if (!pc->chan->signalmode) {
#else
	if (!0) {
#endif
		pc->para.cause = 50; /* facility not subscribed */
		l3dss1_nt_msg_resume_reject(pc, pr, NULL);
		return;
	}

	/* Now we are on none mandatory IEs */
	err = nt_check_infoelements(pc, skb, ie_RESUME);
	if (ERR_IE_COMPREHENSION == err) {
		pc->para.cause = 96;
		l3dss1_nt_msg_release_complete(pc, pr, NULL);
		return;
	}

	/* get bchannel */
	pc->para.bchannel = l3dss1_nt_get_channel_id(pc, skb);
	pc->para.bchannel_inuse = 0;
	if (!l3_get_bchannel(pc,pr)) {
		pc->para.cause = 34; /* no channel */
		l3dss1_nt_msg_resume_reject(pc, pr, NULL);
		return;
	}
	pc->para.bchannel_inuse = 1;

	/* get call identity and store it in setup.phone */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALL_ID, 0))) {
		memcpy(pc->para.setup.phone, p+2, p[1]);
		pc->para.setup.phone[p[1]] = 0;
	} else {
		pc->para.setup.phone[0] = 0;
	}
	pc->para.setup.plan = 0;
	pc->para.setup.screen = 0;

	newl3state(pc, 17); /* resume request */
	if (err) /* STATUS for none mandatory IE errors after actions are taken */
		l3dss1_nt_std_ie_err(pc, err);
	/* tell call control about setup */
	pc->st->l3.l3l4(pc->st, CC_RESUME | INDICATION, pc);
}

/* received a suspend request from user
 */
static void
l3dss1_nt_suspend(struct l3_process *pc, u_char pr, void *arg)
{
	u_char *p;
	struct sk_buff *skb = arg;
	int err = 0;

	/* only signalmode supports resume */
#ifdef CONFIG_HISAX_SIGNALMODE
	if (!pc->chan->signalmode) {
#else
	if (!0) {
#endif
		pc->para.cause = 50; /* facility not subscribed */
		l3dss1_nt_msg_resume_reject(pc, pr, NULL);
		return;
	}

	/* Now we are on none mandatory IEs */
	err = nt_check_infoelements(pc, skb, ie_SUSPEND);
	if (ERR_IE_COMPREHENSION == err) {
		pc->para.cause = 96;
		l3dss1_nt_msg_release_complete(pc, pr, NULL);
		return;
	}

	/* get call identity and store it in setup.phone */
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALL_ID, 0))) {
		memcpy(pc->para.setup.phone, p+2, p[1]);
		pc->para.setup.phone[p[1]] = 0;
	} else {
		pc->para.setup.phone[0] = 0;
	}
	pc->para.setup.plan = 0;
	pc->para.setup.screen = 0;

	newl3state(pc, 15); /* suspend request */
	if (err) /* STATUS for none mandatory IE errors after actions are taken */
		l3dss1_nt_std_ie_err(pc, err);
	/* tell call control about setup */
	pc->st->l3.l3l4(pc->st, CC_SUSPEND | INDICATION, pc);
}

/* switch into overlap mode. if event is CC_IGNORE | REQUEST, turn on dialtone
 */
static void
l3dss1_nt_overlap(struct l3_process *pc, u_char pr, void *arg)
{
	//struct IsdnCardState *cs = pc->st->l1.hardware;
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
  	struct IsdnCardState *cs = pc->st->l1.hardware;


	/* this routine is called if there is more dialing information to come from the te. */

	/* the T302 will be started */
	if (cs->nt.dialtone_tout)
		L3AddTimer(&pc->timer, cs->nt.dialtone_tout, CC_T302);

	/* play own tones if not in signal mode */
	if (!pc->chan->signalmode)
		if (pc->para.bchannel_inuse && pc->para.setup.si1==1) /* only if channel is in use and the service indicate is set to audio */
			bchannel_nt_allow_tones(cs, pc->para.bchannel);

	/* reply with setup acknowledge */
	MsgHead(p, pc->callref, MT_SETUP_ACKNOWLEDGE);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_SETUP_ACKNOWLEDGE, 0, IE_CHANNEL_ID-1);

#endif
	*p++ = IE_CHANNEL_ID;
	*p++ = 1; /* length=1 */
	*p++ = 0x88+pc->para.bchannel;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_SETUP_ACKNOWLEDGE, IE_CHANNEL_ID+1, IE_PROGRESS-1);

#endif
	*p++ = IE_PROGRESS;
	*p++ = 2; /* length=2 */
	*p++ = 0x81; /* private network serving local user */
	*p++ = 0x88; /* dialtone available */

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_SETUP_ACKNOWLEDGE, IE_PROGRESS+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* change state to overlap sending */
	newl3state(pc, 2);

	/* start dial tone*/
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_DIALTONE);

}

/* is called when the connection should be disconnected. a busy tone will be sent */

static void
l3dss1_nt_disconnect_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	u_char cause = 16;
  	struct IsdnCardState *cs = pc->st->l1.hardware;

	if (pc->para.cause != NO_CAUSE)
		cause = pc->para.cause;

	/* start tone*/
	if (pc->para.bchannel_inuse) {
		if (cause==16)
			tones_nt_set(pc,BCH_NT_TONE_BUSY);
		else	tones_nt_set(pc,BCH_NT_TONE_RELEASE);
	}

	/* stop all timers */
	StopAllL3Timer(pc);

	/* send disconnect */
	MsgHead(p, pc->callref, MT_DISCONNECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_DISCONNECT, 0, IE_PROGRESS-1);

#endif
	*p++ = IE_PROGRESS;
	*p++ = 0x2;
	*p++ = 0x81; /* private net serving local user */
	*p++ = 0x88; /* inband info */

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_DISCONNECT, IE_PROGRESS+1, 255);

	if (!(pc->chan->signalmode)) {
#endif
	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

        if (pc->prot.dss1_nt.uus1_data[0]) {
		*p++ = IE_DISPLAY;
	        *p++ = strlen(pc->prot.dss1_nt.uus1_data);
	        strcpy(p,pc->prot.dss1_nt.uus1_data);
	        p += strlen(pc->prot.dss1_nt.uus1_data);
	        pc->prot.dss1_nt.uus1_data[0] = '\0';   
	}

#ifdef CONFIG_HISAX_SIGNALMODE
	}
#endif

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* start timer T306 */
	if (cs->nt.disconn_tout)
		L3AddTimer(&pc->timer, cs->nt.disconn_tout, CC_T306);

	/* change to disconnect_indication state */
	newl3state(pc, 12);

// stupid shit down here:::
//	/* send a timeout to the application */
//#ifdef CONFIG_HISAX_SIGNALMODE
//	memset(&pc->para.signal, 0, sizeof(signal_msg));
//	pc->para.signal.cause = 34; /* no channel */
//	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
//#endif
}


/* any state exept 0,11,12,19,22 will cause this disconnect on request */

static void
l3dss1_nt_disconnect(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	u_char *p;
	int ret;
	u_char cause = 0;

	/* stop all timers */
	StopAllL3Timer(pc);

	if ((ret = l3dss1_nt_get_cause(pc, skb))) {
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "DISC get_cause ret(%d)", ret);
		if (ret < 0)
			cause = 96;
		else if (ret > 0)
			cause = 100;
	} 
	if ((p = findie(skb->data, skb->len, IE_FACILITY, 0)))
		l3dss1_nt_parse_facility(pc->st, pc, pc->callref, p);
	ret = nt_check_infoelements(pc, skb, ie_DISCONNECT);
	if (ERR_IE_COMPREHENSION == ret)
		cause = 96;
	else if ((!cause) && (ERR_IE_UNRECOGNIZED == ret))
		cause = 99;

	/* disconnect b-channel (if not already disconnected) */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_NONE);

	/* new state */
	newl3state(pc, 11);

	/* indicate disconnect */
	pc->st->l3.l3l4(pc->st, CC_DISCONNECT | INDICATION, pc);

	/* hack: no respone from call control expected, so we just continue
	as the call control would answer. this is essential if we are
	in dialing state where we do not have an association to l4. */
       	l3dss1_nt_release_req(pc, pr, arg);
}


/* this disconnect is called when call control will respond to a disconnect indication */
static void
l3dss1_nt_disconnect_2(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	u_char *p;
	int ret;
	u_char cause = 0;

	/* stop all timers */
	StopAllL3Timer(pc);

	if ((ret = l3dss1_nt_get_cause(pc, skb))) {
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "DISC get_cause ret(%d)", ret);
		if (ret < 0)
			cause = 96;
		else if (ret > 0)
			cause = 100;
	} 
	if ((p = findie(skb->data, skb->len, IE_FACILITY, 0)))
		l3dss1_nt_parse_facility(pc->st, pc, pc->callref, p);
	ret = nt_check_infoelements(pc, skb, ie_DISCONNECT);
	if (ERR_IE_COMPREHENSION == ret)
		cause = 96;
	else if ((!cause) && (ERR_IE_UNRECOGNIZED == ret))
		cause = 99;

	/* disconnect b-channel (if not already disconnected) */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc, BCH_NT_TONE_NONE);

	/* new state */
	newl3state(pc, 11);

	/* indicate disconnect */
	pc->st->l3.l3l4(pc->st, CC_DISCONNECT | INDICATION, pc);
}


/* a setup response is received from call control
*/
static void
l3dss1_nt_setup_rsp4(struct l3_process *pc, u_char pr,
		 void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
	struct timeval now;
	struct rtc_time tm;
  	struct IsdnCardState *cs = pc->st->l1.hardware;
/*        if (!pc->para.bchannel) 
	 { if (pc->debug & L3_DEB_WARN)
	       l3_debug(pc->st, "D-chan connect for waiting call");
           l3dss1_nt_disconnect_req(pc, pr, arg);
           return;
         }
*/
	/* stop tones */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_NONE);

	/* maybe there is a timer running (not standard) */
	L3DelTimer(&pc->timer);

	/* reply with connect */
	MsgHead(p, pc->callref, MT_CONNECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, 0, IE_DATE-1);

#endif
	*p++ = IE_DATE;
	*p++ = 6;
	do_gettimeofday(&now);
	l3dss1_nt_to_tm(now.tv_sec+cs->nt.date_offset, &tm);
	*p++ = (tm.tm_year-100)%100;
	*p++ = tm.tm_mon;
	*p++ = tm.tm_mday;
	*p++ = tm.tm_hour;
	*p++ = tm.tm_min;
	*p++ = tm.tm_sec;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, IE_DATE+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* changing state to active */
	newl3state(pc, 10);

	/* send connect complete to call control */
	pc->chan->l2_protocol=ISDN_PROTO_L2_TRANS;
// jolly: a hack ?? dunno the right way yet..
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
#endif
	pc->st->l3.l3l4(pc->st, CC_SETUP_COMPL | INDICATION, pc);

}


/* a setup response is received from call control
*/
static void
l3dss1_nt_setup_rsp2(struct l3_process *pc, u_char pr,
		 void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
	struct timeval now;
	struct rtc_time tm;
  	struct IsdnCardState *cs = pc->st->l1.hardware;
/*        if (!pc->para.bchannel) 
	 { if (pc->debug & L3_DEB_WARN)
	       l3_debug(pc->st, "D-chan connect for waiting call");
           l3dss1_nt_disconnect_req(pc, pr, arg);
           return;
         }
*/
	/* stop tones */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_NONE);

	/* maybe there is a timer running (not standard) */
	L3DelTimer(&pc->timer);

	/* reply with connect */
	MsgHead(p, pc->callref, MT_CONNECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, 0, IE_DATE-1);

#endif
	*p++ = IE_DATE;
	*p++ = 6;
	do_gettimeofday(&now);
	l3dss1_nt_to_tm(now.tv_sec+cs->nt.date_offset, &tm);
	*p++ = (tm.tm_year-100)%100;
	*p++ = tm.tm_mon;
	*p++ = tm.tm_mday;
	*p++ = tm.tm_hour;
	*p++ = tm.tm_min;
	*p++ = tm.tm_sec;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, IE_DATE+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* in case we have had a tone active, we need to wait for 
	releasing the bchannel */
//	if (pc->chan->b_st->l1.l1m.state == 2) {
//		/* change to state ACTIVE_AFTER_BC_RELEASE */
//		newl3state(pc, 26);
//		return;
//	}

	/* changing state to active */
	newl3state(pc, 10);

	/* send connect complete to call control */
	pc->chan->l2_protocol=ISDN_PROTO_L2_TRANS;
	// jolly: a hack ?? dunno the right way yet..
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
#endif
	pc->st->l3.l3l4(pc->st, CC_SETUP_COMPL | INDICATION, pc);
}


/* a setup response is received from call control
*/
static void
l3dss1_nt_setup_rsp1(struct l3_process *pc, u_char pr,
		 void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
	struct timeval now;
	struct rtc_time tm;
  	struct IsdnCardState *cs = pc->st->l1.hardware;

/*        if (!pc->para.bchannel) 
	 { if (pc->debug & L3_DEB_WARN)
	       l3_debug(pc->st, "D-chan connect for waiting call");
           l3dss1_nt_disconnect_req(pc, pr, arg);
           return;
         }
*/

	/* Channel Identification */
//	if (!l3_get_bchannel(pc,pr)) /* tones are blocked */
//		return;
//	pc->para.bchannel_inuse = 1;

	/* reply with connect */
	MsgHead(p, pc->callref, MT_CONNECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, 0, IE_CHANNEL_ID-1);

#endif
	*p++ = IE_CHANNEL_ID;
	*p++ = 1; /* length=1 */
	*p++ = 0x88+pc->para.bchannel;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, IE_CHANNEL_ID+1, IE_PROGRESS-1);

#endif
	*p++ = IE_PROGRESS;
	*p++ = 2; /* length=2 */
	*p++ = 0x81; /* private network serving local user */
	*p++ = 0x88; /* dialtone available */

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, IE_PROGRESS+1, IE_DATE-1);

#endif
	*p++ = IE_DATE;
	*p++ = 6;
	do_gettimeofday(&now);
	l3dss1_nt_to_tm(now.tv_sec+cs->nt.date_offset, &tm);
	*p++ = (tm.tm_year-100)%100;
	*p++ = tm.tm_mon;
	*p++ = tm.tm_mday;
	*p++ = tm.tm_hour;
	*p++ = tm.tm_min;
	*p++ = tm.tm_sec;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT, IE_DATE+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* maybe there is a timer running (not standard) */
	L3DelTimer(&pc->timer);

	/* changing state to active */
	newl3state(pc, 10);

	/* send connect complete to call control */
	pc->chan->l2_protocol=ISDN_PROTO_L2_TRANS;
	// jolly: a hack ?? dunno the right way yet..
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
#endif
	pc->st->l3.l3l4(pc->st, CC_SETUP_COMPL | INDICATION, pc);
}

static void
l3dss1_nt_reject_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	u_char cause = 21;

	if (pc->para.cause != NO_CAUSE)
		cause = pc->para.cause;

	MsgHead(p, pc->callref, MT_RELEASE_COMPLETE);

	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);
	newl3state(pc, 0);
	dss1_nt_release_l3_process(pc);
}


/* acknowledge the suspend to user, the call is released */
static void
l3dss1_nt_suspend_ack(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, MT_SUSPEND_ACKNOWLEDGE);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_SUSPEND_ACKNOWLEDGE, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	newl3state(pc, 0);
	dss1_nt_release_l3_process(pc);
}


/* reject suspend to user, the call stays active */
static void
l3dss1_nt_suspend_rej(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	u_char cause = 21;

	MsgHead(p, pc->callref, MT_SUSPEND_REJECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_SUSPEND_REJECT, 0, 255);
	else { /**/
#endif
	if (pc->para.cause != NO_CAUSE)
		cause = pc->para.cause;

#ifdef CONFIG_HISAX_SIGNALMODE
	} /**/
#endif
	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	newl3state(pc, 10);
}


/* acknowledge resume to user, the call become active */
static void
l3dss1_nt_resume_ack(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, MT_RESUME_ACKNOWLEDGE);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_RESUME_ACKNOWLEDGE, 0, IE_CHANNEL_ID-1);

#endif
	*p++ = IE_CHANNEL_ID;
	*p++ = 1; /* length=1 */
	*p++ = 0x88+pc->para.bchannel;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_RESUME_ACKNOWLEDGE, IE_CHANNEL_ID+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	newl3state(pc, 10);
}

/* reject resume, the call stays idle */
static void
l3dss1_nt_resume_rej(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	u_char cause = 21;

	MsgHead(p, pc->callref, MT_RESUME_REJECT);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_RESUME_REJECT, 0, 255);
	else { /**/
#endif
	if (pc->para.cause != NO_CAUSE)
		cause = pc->para.cause;

#ifdef CONFIG_HISAX_SIGNALMODE
	} /**/
#endif
	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	newl3state(pc, 0);
	dss1_nt_release_l3_process(pc);
}

/* an connect acknowledge has been received from call
 */
static void
l3dss1_nt_setup_cpl(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;

	/* send connect acknowledge */
	MsgHead(p, pc->callref, MT_CONNECT_ACKNOWLEDGE);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CONNECT_ACKNOWLEDGE, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* go into active state */
	newl3state(pc, 10);
}

/* we want to send a notify
 */
static void
l3dss1_nt_notify_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, MT_NOTIFY);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_NOTIFY, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

/* we want to send a progress
 */
static void
l3dss1_nt_progress_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
	int l;

	MsgHead(p, pc->callref, MT_PROGRESS);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_PROGRESS, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

/* we want to send information to terminal
 */
static void
l3dss1_nt_information_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
  	struct IsdnCardState *cs = pc->st->l1.hardware;
	int l;

	/* restart timer T304 */
	if (pc->state == 25) {
		/* stop T304 */
		L3DelTimer(&pc->timer);

		/* start T304 */
		if (cs->nt.outoverlap_tout)
			L3AddTimer(&pc->timer, cs->nt.outoverlap_tout, CC_T304);
	}

	MsgHead(p, pc->callref, MT_INFORMATION);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_INFORMATION, 0, 255);
	//else
#endif
	// to be done: setup.phone must be dialed, if needed somewhen


	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

/* we get a release from many states
*/
static void
l3dss1_nt_release(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	u_char *p;
	int ret, cause=0;

	/* stop all timers */
	StopAllL3Timer(pc);

	/* process cause and send release complete */
	if ((ret = l3dss1_nt_get_cause(pc, skb))>0) {
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "REL get_cause ret(%d)", ret);
	} else if (ret<0)
		pc->para.cause = NO_CAUSE;
	if ((p = findie(skb->data, skb->len, IE_FACILITY, 0))) {
		l3dss1_nt_parse_facility(pc->st, pc, pc->callref, p);
	}
	if ((ret<0) && (pc->state != 11))
		cause = 96;
	else if (ret>0)
		cause = 100;
	ret = nt_check_infoelements(pc, skb, ie_RELEASE);
	if (ERR_IE_COMPREHENSION == ret)
		cause = 96;
	else if ((ERR_IE_UNRECOGNIZED == ret) && (!cause))
		cause = 99;  
	if (cause)
		l3dss1_nt_message_cause(pc, MT_RELEASE_COMPLETE, cause);
	else
		l3dss1_nt_message(pc, MT_RELEASE_COMPLETE);

	/* release l4 */
	pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);

	/* disconnect b-channel */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_NONE);

	/* in case we have had a tone active, we need to wait for 
	releasing the bchannel */
//debugl1(pc->st->l1.hardware,"der state: %d",pc->chan->b_st->l1.l1m.state);
//	if (!(pc->para.bchannel_inuse) && pc->chan->b_st->l1.l1m.state == 2) {
//		/* change to state NULL_AFTER_BC_RELEASE */
//		newl3state(pc, 27);
//		return;
//	}

	/* state NULL */
	newl3state(pc, 0);

	/* release */
	dss1_nt_release_l3_process(pc);
}

static void
l3dss1_nt_alert_req(struct l3_process *pc, u_char pr,
		 void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
  	struct IsdnCardState *cs = pc->st->l1.hardware;

	/* stop all timers */
	StopAllL3Timer(pc);

	/* start timer Toutalert */
	if (cs->nt.inalert_tout)
		L3AddTimer(&pc->timer, cs->nt.inalert_tout, CC_Tinalert);

	/* ringing tone */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_RINGING);

	MsgHead(p, pc->callref, MT_ALERTING);

	/* send alerting message to user */
#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_ALERTING, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* change to state 4 */
	newl3state(pc, 4);
}

/* alerting at state 1
we go directly to state 4, because we received an alerting at state 1
*/
static void
l3dss1_nt_alert_req1(struct l3_process *pc, u_char pr,
		 void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
  	struct IsdnCardState *cs = pc->st->l1.hardware;


	/* stop all timers */
	StopAllL3Timer(pc);

	/* start timer Toutalert */
	if (cs->nt.inalert_tout)
		L3AddTimer(&pc->timer, cs->nt.inalert_tout, CC_Tinalert);

	/* Channel Identification */
//	if (!l3_get_bchannel(pc,pr)) /* tones are blocked */
//		return;
//	pc->para.bchannel_inuse = 1;

	/* reply with setup acknowledge */
	MsgHead(p, pc->callref, MT_ALERTING);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_ALERTING, 0, IE_CHANNEL_ID-1);

#endif
	*p++ = IE_CHANNEL_ID;
	*p++ = 1; /* length=1 */
	*p++ = 0x88+pc->para.bchannel;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_ALERTING, IE_CHANNEL_ID+1, IE_PROGRESS-1);

#endif
	*p++ = IE_PROGRESS;
	*p++ = 2; /* length=2 */
	*p++ = 0x81; /* private network serving local user */
	*p++ = 0x88; /* dialtone available */

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_ALERTING, IE_PROGRESS+1, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);

	/* change to state 4 */
	newl3state(pc, 4);
}

/* a proceeding signal was sent by callc
 */
static void
l3dss1_nt_proceed_req(struct l3_process *pc, u_char pr,
		   void *arg)
{
	u_char tmp[256];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
  	struct IsdnCardState *cs = pc->st->l1.hardware;

	/* stop timer T302 */
	StopAllL3Timer(pc);

	/* start timer Tinproc */
	if (cs->nt.inproc_tout)
		L3AddTimer(&pc->timer, cs->nt.inproc_tout, CC_Tinproc);

	/* turn of dialtone sound, if set */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_SILENCE);

	/* change to proceeding state */
	newl3state(pc, 3);

	/* send proceeding message */
	MsgHead(p, pc->callref, MT_CALL_PROCEEDING);

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode)
		p = put_setupextensions(pc, p, MT_CALL_PROCEEDING, 0, 255);

#endif
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}

#if 0
static void
l3dss1_nt_setup_ack_req(struct l3_process *pc, u_char pr,
		   void *arg)
{
	newl3state(pc, 25);
	L3DelTimer(&pc->timer);
	if (cs->nt.overlap_tout)
		L3AddTimer(&pc->timer, cs->nt.overlap_tout, CC_T302);
	l3dss1_nt_message(pc, MT_SETUP_ACKNOWLEDGE);
}
#endif

#if 0
/********************************************/
/* deliver a incoming display message to HL */
/********************************************/
static void
l3dss1_nt_deliver_display(struct l3_process *pc, int pr, u_char *infp)
{       u_char len;
        isdn_ctrl ic; 
	struct IsdnCardState *cs;
        char *p; 

        if (*infp++ != IE_DISPLAY) return;
        if ((len = *infp++) > 80) return; /* total length <= 82 */
	if (!pc->chan) return;

	p = ic.parm.display; 
        while (len--)
	  *p++ = *infp++;
	*p = '\0';
	ic.command = ISDN_STAT_DISPLAY;
	cs = pc->st->l1.hardware;
	ic.driver = cs->myid;
	ic.arg = pc->chan->chan; 
	cs->iif.statcallb(&ic);
} /* l3dss1_nt_deliver_display */


#endif

static void
l3dss1_nt_progress(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	int err = 0;
	u_char *p;

	if ((p = findie(skb->data, skb->len, IE_PROGRESS, 0))) {
		if (p[1] != 2) {
			err = 1;
			pc->para.cause = 100;
		} else if (!(p[2] & 0x70)) {
			switch (p[2]) {
				case 0x80:
				case 0x81:
				case 0x82:
				case 0x84:
				case 0x85:
				case 0x87:
				case 0x8a:
					switch (p[3]) {
						case 0x81:
						case 0x82:
						case 0x83:
						case 0x84:
						case 0x88:
							break;
						default:
							err = 2;
							pc->para.cause = 100;
							break;
					}
					break;
				default:
					err = 3;
					pc->para.cause = 100;
					break;
			}
		}
	} else {
		pc->para.cause = 96;
		err = 4;
	}
	if (err) {	
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "progress error %d", err);
		l3dss1_nt_status_send(pc, pr, NULL);
		return;
	}
	/* Now we are on none mandatory IEs */
	err = nt_check_infoelements(pc, skb, ie_PROGRESS);
	if (err)
		l3dss1_nt_std_ie_err(pc, err);
	if (ERR_IE_COMPREHENSION != err)
		pc->st->l3.l3l4(pc->st, CC_PROGRESS | INDICATION, pc);
}


/* we received a notify
 */
static void
l3dss1_nt_notify(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	int err = 0;
	u_char *p;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (!(pc->chan->signalmode)) { /**/
#endif
	if ((p = findie(skb->data, skb->len, IE_NOTIFY, 0))) {
		if (p[1] != 1) {
			err = 1;
			pc->para.cause = 100;
		} else {
			switch (p[2]) {
				case 0x80:
				case 0x81:
				case 0x82:
					break;
				default:
					pc->para.cause = 100;
					err = 2;
					break;
			}
		}
	} else {
		pc->para.cause = 96;
		err = 3;
	}
	if (err) {	
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "notify error %d", err);
		l3dss1_nt_status_send(pc, pr, NULL);
		return;
	}
	/* Now we are on none mandatory IEs */
	err = nt_check_infoelements(pc, skb, ie_NOTIFY);
	if (err)
		l3dss1_nt_std_ie_err(pc, err);
#ifdef CONFIG_HISAX_SIGNALMODE
	} /**/
#endif
	if (ERR_IE_COMPREHENSION != err)
		pc->st->l3.l3l4(pc->st, CC_NOTIFY | INDICATION, pc);
}

#if 0
static void
l3dss1_nt_status_enq(struct l3_process *pc, u_char pr, void *arg)
{
	int ret;
	struct sk_buff *skb = arg;

	ret = nt_check_infoelements(pc, skb, ie_STATUS_ENQUIRY);
	l3dss1_nt_std_ie_err(pc, ret);
	pc->para.cause = 30; /* response to STATUS_ENQUIRY */
        l3dss1_nt_status_send(pc, pr, NULL);
}
#endif

/* additional dialing information is received in state 2 
 * in other states these are received keypad stuff
 */
static void
l3dss1_nt_information(struct l3_process *pc, u_char pr, void *arg)
{
	int ret;
	struct sk_buff *skb = arg;
	u_char *p;
	char tmp[256];
  	struct IsdnCardState *cs = pc->st->l1.hardware;

#ifdef CONFIG_HISAX_SIGNALMODE
	if (pc->chan->signalmode) {
		/* reset time if overlap sending */
		if (pc->state == 2) { /* overlap sending */
			if (cs->nt.overlap_tout)
				L3AddTimer(&pc->timer, cs->nt.overlap_tout, CC_T302);
		}
		/* send shit to callc */
		pc->st->l3.l3l4(pc->st, CC_INFORMATION | INDICATION, pc);
		/* done */
		return;
	}
#endif
	ret = nt_check_infoelements(pc, skb, ie_INFORMATION);
	if (ret)
		l3dss1_nt_std_ie_err(pc, ret);
//	if (pc->state == 2) { /* overlap sending */
	if (pc->para.bchannel_inuse)
		tones_nt_set(pc,BCH_NT_TONE_SILENCE);
	L3DelTimer(&pc->timer);
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CALLED_PN, 0))) {
		iecpy(tmp, p, 1);
		if (strlen(pc->para.setup.eazmsn)+strlen(tmp)+1 >= sizeof (pc->para.setup.eazmsn)) {
			pc->para.loc = 0;
			pc->para.cause = 28; // invalid number 
			l3dss1_nt_disconnect_req(pc, pr, NULL);
			return;
		}
		strcat(pc->para.setup.eazmsn, tmp);
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_KEYPAD, 0))) {
		iecpy(tmp, p, 0);
		if (strlen(pc->para.setup.eazmsn)+strlen(tmp)+1 >= sizeof (pc->para.setup.eazmsn)) {
			pc->para.loc = 0;
			pc->para.cause = 28; // invalid number 
			l3dss1_nt_disconnect_req(pc, pr, NULL);
			return;
		}
		strcat(pc->para.setup.eazmsn, tmp);
	}

	/* overlap timeout */
	if (pc->state == 2) { /* overlap sending */
		if (cs->nt.overlap_tout)
			L3AddTimer(&pc->timer, cs->nt.overlap_tout, CC_T302);
		pc->st->l3.l3l4(pc->st, CC_INFORMATION | INDICATION, pc);
	}
}

#if 0
/******************************/
/* handle deflection requests */
/******************************/
static void l3dss1_nt_redir_req(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb;
	u_char tmp[256];
	u_char *p = tmp;
        u_char *subp;
        u_char len_phone = 0;
        u_char len_sub = 0;
	int l; 


        strcpy(pc->prot.dss1_nt.uus1_data,pc->chan->setup.eazmsn); /* copy uus element if available */
        if (!pc->chan->setup.phone[0])
          { pc->para.cause = -1;
            l3dss1_nt_disconnect_req(pc,pr,arg); /* disconnect immediately */
            return;
          } /* only uus */
 
        if (pc->prot.dss1_nt.invoke_id) 
          nt_free_invoke_id(pc->st,pc->prot.dss1_nt.invoke_id);
 
        if (!(pc->prot.dss1_nt.invoke_id = nt_new_invoke_id(pc->st))) 
          return;

        MsgHead(p, pc->callref, MT_FACILITY);

        for (subp = pc->chan->setup.phone; (*subp) && (*subp != '.'); subp++) len_phone++; /* len of phone number */
        if (*subp++ == '.') len_sub = strlen(subp) + 2; /* length including info subaddress element */ 

	*p++ = 0x1c;   /* Facility info element */
        *p++ = len_phone + len_sub + 2 + 2 + 8 + 3 + 3; /* length of element */
        *p++ = 0x91;  /* remote operations protocol */
        *p++ = 0xa1;  /* invoke component */
	  
        *p++ = len_phone + len_sub + 2 + 2 + 8 + 3; /* length of data */
        *p++ = 0x02;  /* invoke id tag, integer */
	*p++ = 0x01;  /* length */
        *p++ = pc->prot.dss1_nt.invoke_id;  /* invoke id */ 
        *p++ = 0x02;  /* operation value tag, integer */
	*p++ = 0x01;  /* length */
        *p++ = 0x0D;  /* Call Deflect */
	  
        *p++ = 0x30;  /* sequence phone number */
        *p++ = len_phone + 2 + 2 + 3 + len_sub; /* length */
	  
        *p++ = 0x30;  /* Deflected to UserNumber */
        *p++ = len_phone+2+len_sub; /* length */
        *p++ = 0x80; /* NumberDigits */
	*p++ = len_phone; /* length */
        for (l = 0; l < len_phone; l++)
	 *p++ = pc->chan->setup.phone[l];

        if (len_sub)
	  { *p++ = 0x04; /* called party subaddress */
            *p++ = len_sub - 2;
            while (*subp) *p++ = *subp++;
          }

        *p++ = 0x01; /* screening identifier */
        *p++ = 0x01;
        *p++ = pc->chan->setup.screen;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l))) return;
	memcpy(skb_put(skb, l), tmp, l);

        l3_msg(pc->st, DL_DATA | REQUEST, skb);
} /* l3dss1_nt_redir_req */

/********************************************/
/* handle deflection request in early state */
/********************************************/
static void l3dss1_nt_redir_req_early(struct l3_process *pc, u_char pr, void *arg)
{
  l3dss1_nt_proceed_req(pc,pr,arg);
  l3dss1_nt_redir_req(pc,pr,arg);
} /* l3dss1_nt_redir_req_early */
#endif

/***********************************************/
/* handle special commands for this protocol.  */
/* Examples are call independant services like */
/* remote operations with dummy  callref.      */
/***********************************************/
static int l3dss1_nt_cmd_global(struct PStack *st, isdn_ctrl *ic)
{ u_char id;
  u_char temp[265];
  u_char *p = temp;
  int i, l, proc_len; 
  struct sk_buff *skb;
  struct l3_process *pc = NULL;

  switch (ic->arg)
   { case DSS1_CMD_INVOKE:
       if (ic->parm.dss1_io.datalen < 0) return(-2); /* invalid parameter */ 

       for (proc_len = 1, i = ic->parm.dss1_io.proc >> 8; i; i++) 
         i = i >> 8; /* add one byte */    
       l = ic->parm.dss1_io.datalen + proc_len + 8; /* length excluding ie header */
       if (l > 255) 
         return(-2); /* too long */

       if (!(id = nt_new_invoke_id(st))) 
         return(0); /* first get a invoke id -> return if no available */
       
       i = -1; 
       MsgHead(p, i, MT_FACILITY); /* build message head */
       *p++ = 0x1C; /* Facility IE */
       *p++ = l; /* length of ie */
       *p++ = 0x91; /* remote operations */
       *p++ = 0xA1; /* invoke */
       *p++ = l - 3; /* length of invoke */
       *p++ = 0x02; /* invoke id tag */
       *p++ = 0x01; /* length is 1 */
       *p++ = id; /* invoke id */
       *p++ = 0x02; /* operation */
       *p++ = proc_len; /* length of operation */
       
       for (i = proc_len; i; i--)
         *p++ = (ic->parm.dss1_io.proc >> (i-1)) & 0xFF;
       memcpy(p, ic->parm.dss1_io.data, ic->parm.dss1_io.datalen); /* copy data */
       l = (p - temp) + ic->parm.dss1_io.datalen; /* total length */         

       if (ic->parm.dss1_io.timeout > 0)
        if (!(pc = dss1_nt_new_l3_process(st, -1, 0)))
          { nt_free_invoke_id(st, id);
            return(-2);
          } 
       pc->prot.dss1_nt.ll_id = ic->parm.dss1_io.ll_id; /* remember id */ 
       pc->prot.dss1_nt.proc = ic->parm.dss1_io.proc; /* and procedure */

       if (!(skb = l3_alloc_skb(l))) 
         { nt_free_invoke_id(st, id);
           if (pc) dss1_nt_release_l3_process(pc);
           return(-2);
         }
       memcpy(skb_put(skb, l), temp, l);
       
       if (pc)
        { pc->prot.dss1_nt.invoke_id = id; /* remember id */
          L3AddTimer(&pc->timer, ic->parm.dss1_io.timeout, CC_TDSS1_IO | REQUEST);
        }
       
       l3_msg(st, DL_DATA | REQUEST, skb);
       ic->parm.dss1_io.hl_id = id; /* return id */
       return(0);

     case DSS1_CMD_INVOKE_ABORT:
       if ((pc = l3dss1_nt_search_dummy_proc(st, ic->parm.dss1_io.hl_id)))
	{ L3DelTimer(&pc->timer); /* remove timer */
          dss1_nt_release_l3_process(pc);
          return(0); 
        } 
       else
	{ l3_debug(st, "l3dss1_nt_cmd_global abort unknown id");
          return(-2);
        } 
       break;
    
     default: 
       l3_debug(st, "l3dss1_nt_cmd_global unknown cmd 0x%lx", ic->arg);
       return(-1);  
   } /* switch ic-> arg */
  return(-1);
} /* l3dss1_nt_cmd_global */

static void 
l3dss1_nt_io_timer(struct l3_process *pc)
{ isdn_ctrl ic;
  struct IsdnCardState *cs = pc->st->l1.hardware;

  L3DelTimer(&pc->timer); /* remove timer */

  ic.driver = cs->myid;
  ic.command = ISDN_STAT_PROT;
  ic.arg = DSS1_STAT_INVOKE_ERR;
  ic.parm.dss1_io.hl_id = pc->prot.dss1_nt.invoke_id;
  ic.parm.dss1_io.ll_id = pc->prot.dss1_nt.ll_id;
  ic.parm.dss1_io.proc = pc->prot.dss1_nt.proc;
  ic.parm.dss1_io.timeout= -1;
  ic.parm.dss1_io.datalen = 0;
  ic.parm.dss1_io.data = NULL;
  nt_free_invoke_id(pc->st, pc->prot.dss1_nt.invoke_id);
  pc->prot.dss1_nt.invoke_id = 0; /* reset id */

  cs->iif.statcallb(&ic);

  dss1_nt_release_l3_process(pc); 
} /* l3dss1_nt_io_timer */

#if 0
static void
l3dss1_nt_release_ind(struct l3_process *pc, u_char pr, void *arg)
{
	u_char *p;
	struct sk_buff *skb = arg;
	int callState = 0;
	p = skb->data;

	if ((p = findie(p, skb->len, IE_CALL_STATE, 0))) {
		p++;
		if (1 == *p++)
			callState = *p;
	}
	if (callState == 0) {
		/* ETS 300-104 7.6.1, 8.6.1, 10.6.1... and 16.1
		 * set down layer 3 without sending any message
		 */
		pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);
		newl3state(pc, 0);
		dss1_nt_release_l3_process(pc);
	} else {
		pc->st->l3.l3l4(pc->st, CC_IGNORE | INDICATION, pc);
	}
}
#endif


/* do absolutely nothing */
static void
l3dss1_nt_dummy(struct l3_process *pc, u_char pr, void *arg)
{
}


/* state 2: timeout of timer T302 */
static void
l3dss1_nt_t302(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.loc = 0;
	pc->para.cause = 28; // invalid number 
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->para.signal.cause = 28;
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	l3dss1_nt_disconnect_req(pc, pr, NULL);
	pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);
}

static void
l3dss1_nt_t306(struct l3_process *pc, u_char pr, void *arg)
{
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	l3dss1_nt_release_req(pc, pr, arg);
}

/* timeout after broadcasting setup with no reply
 */
static void
l3dss1_nt_t303(struct l3_process *pc, u_char pr, void *arg)
{
	if (pc->N303 > 0) {
		pc->N303--;
		/* stop timer */
		L3DelTimer(&pc->timer);

		/* send setup again and start timer */
		l3dss1_nt_setup_req_cont(pc, pr, arg);
	} else {
		/* stop timer */
		L3DelTimer(&pc->timer);

		/* send release complete */
		l3dss1_nt_message_cause(pc, MT_RELEASE_COMPLETE, 102);

		/* set cause */
		pc->para.cause = 18;
#ifdef CONFIG_HISAX_SIGNALMODE
		memset(&pc->para.signal, 0, sizeof(signal_msg));
		pc->para.signal.cause = 18;
#endif

		/* indicate notsetup */
		pc->st->l3.l3l4(pc->st, CC_NOSETUP_RSP, pc);

		/* release cr and process */
		dss1_nt_release_l3_process(pc);
	}
}

static void
l3dss1_nt_t304(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);

	/* set cause */
	pc->para.loc = 0;
	pc->para.cause = 102;
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
	pc->para.signal.cause = 102;
#endif

	l3dss1_nt_disconnect_req(pc, pr, NULL);
	pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);

}

#if 0
static void
l3dss1_nt_t305(struct l3_process *pc, u_char pr, void *arg)
{
	u_char tmp[16];
	u_char *p = tmp;
	int l;
	struct sk_buff *skb;
	u_char cause = 16;

	L3DelTimer(&pc->timer);
	if (pc->para.cause != NO_CAUSE)
		cause = pc->para.cause;

	MsgHead(p, pc->callref, MT_RELEASE);

	*p++ = IE_CAUSE;
	*p++ = 0x2;
	*p++ = 0x81;
	*p++ = cause | 0x80;

	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	newl3state(pc, 19);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
	L3AddTimer(&pc->timer, T308, CC_T308_1);
}

#endif

/* incomming call was proceeding too long */
static void
l3dss1_nt_tinproc(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.loc = 0;
	pc->para.cause = 18;
	l3dss1_nt_disconnect_req(pc, pr, NULL);
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);
}

/* outgoing call was proceeding too long */
static void
l3dss1_nt_toutproc(struct l3_process *pc, u_char pr, void *arg)
{
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	l3dss1_nt_release_req(pc, pr, arg);
}

/* incomming call was alerting too long */
static void
l3dss1_nt_tinalert(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.loc = 0;
	pc->para.cause = 19;
	l3dss1_nt_disconnect_req(pc, pr, NULL);
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);
}

/* outgoing call was alerting too long */
static void
l3dss1_nt_toutalert(struct l3_process *pc, u_char pr, void *arg)
{
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->st->l3.l3l4(pc->st, CC_TIMEOUT | INDICATION, pc);
#endif
	l3dss1_nt_release_req(pc, pr, arg);
}

#if 0
static void
l3dss1_nt_t313(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.loc = 0;
	pc->para.cause = 102;
	l3dss1_nt_disconnect_req(pc, pr, NULL);
	pc->st->l3.l3l4(pc->st, CC_CONNECT_ERR, pc);
}

static void
l3dss1_nt_t308_1(struct l3_process *pc, u_char pr, void *arg)
{
	newl3state(pc, 19);
	L3DelTimer(&pc->timer);
	l3dss1_nt_message(pc, MT_RELEASE);
	L3AddTimer(&pc->timer, T308, CC_T308_2);
}

static void
l3dss1_nt_t308_2(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->st->l3.l3l4(pc->st, CC_RELEASE_ERR, pc);
	dss1_nt_release_l3_process(pc);
}

static void
l3dss1_nt_t318(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.cause = 102;	/* Timer expiry */
	pc->para.loc = 0;	/* local */
	pc->st->l3.l3l4(pc->st, CC_RESUME_ERR, pc);
	newl3state(pc, 19);
	l3dss1_nt_message(pc, MT_RELEASE);
	L3AddTimer(&pc->timer, T308, CC_T308_1);
}

static void
l3dss1_nt_t319(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
	pc->para.cause = 102;	/* Timer expiry */
	pc->para.loc = 0;	/* local */
	pc->st->l3.l3l4(pc->st, CC_SUSPEND_ERR, pc);
	newl3state(pc, 10);
}

static void
l3dss1_nt_restart(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
#endif
	pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);
	dss1_nt_release_l3_process(pc);
}

#endif
static void
l3dss1_nt_status(struct l3_process *pc, u_char pr, void *arg)
{
	u_char *p;
	struct sk_buff *skb = arg;
	int ret; 
	u_char cause = 0, callState = 0;
	
	if ((ret = l3dss1_nt_get_cause(pc, skb))) {
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "STATUS get_cause ret(%d)",ret);
		if (ret < 0)
			cause = 96;
		else if (ret > 0)
			cause = 100;
	}
	if ((p = findie(skb->data, skb->len, IE_CALL_STATE, 0))) {
		p++;
		if (1 == *p++) {
			callState = *p;
			if (!nt_ie_in_set(pc, *p, nt_l3_valid_states))
				cause = 100;
		} else
			cause = 100;
	} else
		cause = 96;
	if (!cause) { /*  no error before */
		ret = nt_check_infoelements(pc, skb, ie_STATUS);
		if (ERR_IE_COMPREHENSION == ret)
			cause = 96;
		else if (ERR_IE_UNRECOGNIZED == ret)
			cause = 99;
	}
	if (cause) {
		u_char tmp;
		
		if (pc->debug & L3_DEB_WARN)
			l3_debug(pc->st, "STATUS error(%d/%d)",ret,cause);
		tmp = pc->para.cause;
		pc->para.cause = cause;
		l3dss1_nt_status_send(pc, 0, NULL);
		if (cause == 99)
			pc->para.cause = tmp;
		else
			return;
	}
	cause = pc->para.cause;
	if (((cause & 0x7f) == 111) && (callState == 0)) {
		/* ETS 300-104 7.6.1, 8.6.1, 10.6.1...
		 * if received MT_STATUS with cause == 111 and call
		 * state == 0, then we must set down layer 3
		 */
#ifdef CONFIG_HISAX_SIGNALMODE
		/* fill info for signal mode extension */
		memset(&pc->para.signal, 0, sizeof(signal_msg));
		pc->para.signal.cause = pc->para.cause;

#endif
		pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);
		newl3state(pc, 0);
		dss1_nt_release_l3_process(pc);
	}
}

#if 0
static void
l3dss1_nt_facility(struct l3_process *pc, u_char pr, void *arg)
{
	struct sk_buff *skb = arg;
	int ret;
	
	ret = nt_check_infoelements(pc, skb, ie_FACILITY);
	l3dss1_nt_std_ie_err(pc, ret);
 	  {
		u_char *p;
		if ((p = findie(skb->data, skb->len, IE_FACILITY, 0)))
			l3dss1_nt_parse_facility(pc->st, pc, pc->callref, p);
	}
}
#endif

static void
l3dss1_nt_global_restart(struct l3_process *pc, u_char pr, void *arg)
{
	u_char tmp[32];
	u_char *p;
	u_char ri, ch = 0, chan = 0;
	int l;
	struct sk_buff *skb = arg;
	struct l3_process *up;

	newl3state(pc, 2);
	L3DelTimer(&pc->timer);
	p = skb->data;
	if ((p = findie(p, skb->len, IE_RESTART_IND, 0))) {
		ri = p[2];
		l3_debug(pc->st, "Restart %x", ri);
	} else {
		l3_debug(pc->st, "Restart without restart IE");
		ri = 0x86;
	}
	p = skb->data;
	if ((p = findie(p, skb->len, IE_CHANNEL_ID, 0))) {
		chan = p[2] & 3;
		ch = p[2];
		if (pc->st->l3.debug)
			l3_debug(pc->st, "Restart for channel %d", chan);
	}
	newl3state(pc, 2);
	up = pc->st->l3.proc;
	while (up) {
		if ((ri & 7) == 7)
			up->st->lli.l4l3(up->st, CC_RESTART | REQUEST, up);
		else if (up->para.bchannel == chan)
			up->st->lli.l4l3(up->st, CC_RESTART | REQUEST, up);
		up = up->next;
	}
	p = tmp;
	MsgHead(p, pc->callref, MT_RESTART_ACKNOWLEDGE);
	if (chan) {
		*p++ = IE_CHANNEL_ID;
		*p++ = 1;
		*p++ = ch | 0x88;
	}
	*p++ = 0x79;		/* RESTART Ind */
	*p++ = 1;
	*p++ = ri;
	l = p - tmp;
	if (!(skb = l3_alloc_skb(l)))
		return;
	memcpy(skb_put(skb, l), tmp, l);
	newl3state(pc, 0);
	l3_msg(pc->st, DL_DATA | REQUEST, skb);
}


static void
l3dss1_nt_dl_reset(struct l3_process *pc, u_char pr, void *arg)
{
        pc->para.cause = 0x29;          /* Temporary failure */
        pc->para.loc = 0;
        l3dss1_nt_disconnect_req(pc, pr, NULL);
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->para.signal.cause = 0x29;
#endif
        pc->st->l3.l3l4(pc->st, CC_SETUP_ERR, pc);
}

static void
l3dss1_nt_dl_release(struct l3_process *pc, u_char pr, void *arg)
{
        newl3state(pc, 0);
        pc->para.cause = 0x1b;          /* Destination out of order */
        pc->para.loc = 0;
#ifdef CONFIG_HISAX_SIGNALMODE
	memset(&pc->para.signal, 0, sizeof(signal_msg));
	pc->para.signal.cause = 0x1b;
#endif
        pc->st->l3.l3l4(pc->st, CC_RELEASE | INDICATION, pc);
        release_l3_process(pc);
}

static void
l3dss1_nt_dl_reestablish(struct l3_process *pc, u_char pr, void *arg)
{
        L3DelTimer(&pc->timer);
        L3AddTimer(&pc->timer, T309, CC_T309);
        l3_msg(pc->st, DL_ESTABLISH | REQUEST, NULL);
}
 
static void
l3dss1_nt_dl_reest_status(struct l3_process *pc, u_char pr, void *arg)
{
	L3DelTimer(&pc->timer);
 
 	pc->para.cause = 0x1F; /* normal, unspecified */
	l3dss1_nt_status_send(pc, 0, NULL);
}

/* *INDENT-OFF* */
/* everything comming from call control */
static struct stateentry downstatelist_nt[] =
{
	/* CC_IGNORE will be handled as CC_MORE_INFO */
	{SBIT(1),
	 CC_IGNORE | REQUEST, l3dss1_nt_overlap},
	{SBIT(1),
	 CC_MORE_INFO | REQUEST, l3dss1_nt_overlap},
//	{SBIT(2),
//	 CC_MORE_INFO | REQUEST, l3dss1_nt_dummy},
	// no overlap receiving yet since this l3 doesn't support ptp
	/* proceeding */
	{SBIT(1) | SBIT(2),
	 CC_PROCEED_SEND | REQUEST, l3dss1_nt_proceed_req},
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) | SBIT(7) | SBIT(8) |
	 SBIT(9) | SBIT(10) | SBIT(11) | SBIT(12) | SBIT(15) | SBIT(17) |
	 SBIT(19) | SBIT(22) | SBIT(25),
	 CC_RELEASE | REQUEST, l3dss1_nt_release_req},
	/* ring after setup */
	{SBIT(1),
	 CC_ALERTING | REQUEST, l3dss1_nt_alert_req1},
	{SBIT(2) | SBIT(3),
	 CC_ALERTING | REQUEST, l3dss1_nt_alert_req},
	{SBIT(10),
	 CC_ALERTING | REQUEST, l3dss1_nt_dummy}, /* this is a hack !! */
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(7) | SBIT(8) | SBIT(10) |
	 SBIT(15) | SBIT(17) |SBIT(25),
         CC_DISCONNECT | REQUEST, l3dss1_nt_disconnect_req},
	{SBIT(8),
	 CC_SETUP_COMPL | REQUEST, l3dss1_nt_setup_cpl},
	/* connect after setup */
	{SBIT(1),
	 CC_SETUP | RESPONSE, l3dss1_nt_setup_rsp1}, /* after initiation */
	{SBIT(2) | SBIT(3),
	 CC_SETUP | RESPONSE, l3dss1_nt_setup_rsp2}, /* after dialing */
	{SBIT(4),
	 CC_SETUP | RESPONSE, l3dss1_nt_setup_rsp4}, /* after ringing */
	{SBIT(1),
	 CC_REJECT | REQUEST, l3dss1_nt_reject_req},
	/* dialing */
	{SBIT(0),
	 CC_SETUP | REQUEST, l3dss1_nt_setup_req},
	/* timers */
	{SBIT(2),
	 CC_T302, l3dss1_nt_t302},
	{SBIT(6),
	 CC_T303, l3dss1_nt_t303},
	{SBIT(25),
	 CC_T304, l3dss1_nt_t304},
	{SBIT(12),
	 CC_T306 | REQUEST, l3dss1_nt_t306},
	{SBIT(10),
	 CC_T309, l3dss1_nt_dl_release},
	{SBIT(3),
	 CC_Tinproc, l3dss1_nt_tinproc},
	{SBIT(4),
	 CC_Tinalert, l3dss1_nt_tinalert},
	{SBIT(7),
	 CC_Toutalert, l3dss1_nt_toutalert},
	{SBIT(9),
	 CC_Toutproc, l3dss1_nt_toutproc},
	/* notify & progress */
	{SBIT(2) | SBIT(3)| SBIT(4)| SBIT(10) | SBIT(11) | SBIT(15),
	 CC_NOTIFY | REQUEST, l3dss1_nt_notify_req},
	{SBIT(2) | SBIT(3) | SBIT(3),
	 CC_PROGRESS | REQUEST, l3dss1_nt_progress_req},
	{SBIT(2) | SBIT(3) | SBIT(4) | SBIT(7) | SBIT(8) | SBIT(9) | SBIT(10) |
	 SBIT(11) | SBIT(12) | SBIT(15) | SBIT(25),
	 CC_INFORMATION | REQUEST, l3dss1_nt_information_req},
	/* suspend & resume */
	{SBIT(15),
	 CC_SUSPEND | RESPONSE, l3dss1_nt_suspend_ack},
	{SBIT(15),
	 CC_SUSPEND_ERR, l3dss1_nt_suspend_rej},
	{SBIT(17),
	 CC_RESUME | RESPONSE, l3dss1_nt_resume_ack},
	{SBIT(17),
	 CC_RESUME_ERR, l3dss1_nt_resume_rej},

//	 CC_T310, l3dss1_nt_t310},
/*
	{SBIT(0),
	 CC_SETUP | REQUEST, l3dss1_nt_setup_req},
	{SBIT(0),
	 CC_RESUME | REQUEST, l3dss1_nt_resume_req},
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) | SBIT(7) | SBIT(8) | SBIT(9) | SBIT(10) | SBIT(25),
	 CC_DISCONNECT | REQUEST, l3dss1_nt_disconnect_req},
	{SBIT(12),
	 CC_RELEASE | REQUEST, l3dss1_nt_release_req},
	{ALL_STATES,
	 CC_RESTART | REQUEST, l3dss1_nt_restart},
	{SBIT(6) | SBIT(25),
	 CC_IGNORE | REQUEST, l3dss1_nt_reset},
	{SBIT(6) | SBIT(25),
	 CC_REJECT | REQUEST, l3dss1_nt_reject_req},
	{SBIT(6) | SBIT(25),
	 CC_PROCEED_SEND | REQUEST, l3dss1_nt_proceed_req},
	{SBIT(6),
	 CC_MORE_INFO | REQUEST, l3dss1_nt_setup_ack_req},
	{SBIT(25),
	 CC_MORE_INFO | REQUEST, l3dss1_nt_dummy},
	{SBIT(6) | SBIT(9) | SBIT(25),
	 CC_ALERTING | REQUEST, l3dss1_nt_alert_req},
	{SBIT(6) | SBIT(7) | SBIT(9) | SBIT(25),
	 CC_SETUP | RESPONSE, l3dss1_nt_setup_rsp},
	{SBIT(10),
	 CC_SUSPEND | REQUEST, l3dss1_nt_suspend_req},
        {SBIT(7) | SBIT(9) | SBIT(25),
         CC_REDIR | REQUEST, l3dss1_nt_redir_req},
        {SBIT(6),
         CC_REDIR | REQUEST, l3dss1_nt_redir_req_early},
        {SBIT(9) | SBIT(25),
         CC_DISCONNECT | REQUEST, l3dss1_nt_disconnect_req},
	{SBIT(25),
	 CC_T302, l3dss1_nt_t302},
	{SBIT(1),
	 CC_T303, l3dss1_nt_t303},
	{SBIT(2),
	 CC_T304, l3dss1_nt_t304},
	{SBIT(3),
	 CC_T310, l3dss1_nt_t310},
	{SBIT(8),
	 CC_T313, l3dss1_nt_t313},
	{SBIT(11),
	 CC_T305, l3dss1_nt_t305},
	{SBIT(15),
	 CC_T319, l3dss1_nt_t319},
	{SBIT(17),
	 CC_T318, l3dss1_nt_t318},
	{SBIT(19),
	 CC_T308_1, l3dss1_nt_t308_1},
	{SBIT(19),
	 CC_T308_2, l3dss1_nt_t308_2},
*/
};

#define DOWNSLLEN \
	(sizeof(downstatelist_nt) / sizeof(struct stateentry))

/* everything comming from user */
// used states: 0,1,2,3,4,6,7,8,9,10,11,12,15,17,19,22,25
static struct stateentry datastatelist_nt[] =
{
	/* incomming setup stuff */
	{SBIT(0),
	 MT_SETUP, l3dss1_nt_setup},
	{SBIT(0),
	 MT_RESUME, l3dss1_nt_resume},
	/* ignore additional setup message because we already received the first one */
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) | SBIT(7) | SBIT(8) |
	 SBIT(9) | SBIT(10) | SBIT(11) | SBIT(12) | SBIT(15) | SBIT(17) |
	 SBIT(19) | SBIT(22) | SBIT(25),
	 MT_SETUP, l3dss1_nt_dummy},
	{SBIT(2) | SBIT(3) | SBIT(4) | SBIT(7) | SBIT(8) | SBIT(9) | SBIT(10) |
	 SBIT(11) | SBIT(12) | SBIT(15) | SBIT(25),
	 MT_INFORMATION, l3dss1_nt_information},
	{SBIT(10),
	 MT_CONNECT_ACKNOWLEDGE, l3dss1_nt_dummy}, /* no message because there is no information from user-te to give */

	/* disconnect and release stuff */
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) | SBIT(7) | SBIT(8) |
	 SBIT(9) | SBIT(10) | SBIT(11) | SBIT(12) | SBIT(15) | SBIT(17) |
	 SBIT(19) | SBIT(22) | SBIT(25),
	 MT_RELEASE, l3dss1_nt_release},
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) | SBIT(7) | SBIT(8) |
	 SBIT(9) | SBIT(10) | SBIT(11) | SBIT(12) | SBIT(15) | SBIT(17) |
	 SBIT(19) | SBIT(22) | SBIT(25),
	 MT_RELEASE_COMPLETE, l3dss1_nt_release_cmpl},
	{SBIT(1) | SBIT(2) | SBIT(3) | SBIT(4) | SBIT(6) |
	 SBIT(15) | SBIT(17) | SBIT(25),
	 MT_DISCONNECT, l3dss1_nt_disconnect},
	{SBIT(7) | SBIT(8) | SBIT (9) | SBIT(10),
	 MT_DISCONNECT, l3dss1_nt_disconnect_2},

	/* outgoing setup stuff */
	{SBIT(6),
	 MT_SETUP_ACKNOWLEDGE, l3dss1_nt_setup_ack},
	{SBIT(6) | SBIT(25),
	 MT_CALL_PROCEEDING, l3dss1_nt_call_proc},
	{SBIT(6) | SBIT(25) | SBIT(9),
	 MT_ALERTING, l3dss1_nt_alerting},
	{SBIT(6) | SBIT(25) | SBIT(7) | SBIT(9),
	 MT_CONNECT, l3dss1_nt_connect},
	/* suspend */
	{SBIT(10),
	 MT_SUSPEND, l3dss1_nt_suspend},
	/* other stuff */
	{SBIT(9) | SBIT(25),
	 MT_PROGRESS, l3dss1_nt_progress},
	{SBIT(2) | SBIT(3) | SBIT(4) | SBIT(10),
	 MT_NOTIFY, l3dss1_nt_notify},
};

#define DATASLLEN \
	(sizeof(datastatelist_nt) / sizeof(struct stateentry))

static struct stateentry globalmes_list_nt[] =
{
	{ALL_STATES,
	 MT_STATUS, l3dss1_nt_status},
	{SBIT(0),
	 MT_RESTART, l3dss1_nt_global_restart},
/*	{SBIT(1),
	 MT_RESTART_ACKNOWLEDGE, l3dss1_nt_restart_ack},
*/
};
#define GLOBALM_LEN \
	(sizeof(globalmes_list_nt) / sizeof(struct stateentry))

static struct stateentry manstatelist_nt[] =
{
	{SBIT(2),
         DL_ESTABLISH | INDICATION, l3dss1_nt_dl_reset},
        {SBIT(10),
         DL_ESTABLISH | CONFIRM, l3dss1_nt_dl_reest_status},
        {SBIT(10),
         DL_RELEASE | INDICATION, l3dss1_nt_dl_reestablish},
        {ALL_STATES,
         DL_RELEASE | INDICATION, l3dss1_nt_dl_release},

};

#define MANSLLEN \
        (sizeof(manstatelist_nt) / sizeof(struct stateentry))
/* *INDENT-ON* */

static void
nt_global_handler(struct PStack *st, int mt, struct sk_buff *skb)
{
	u_char tmp[16];
	u_char *p = tmp;
	int l;
	int i;
	struct l3_process *proc = st->l3.global;

	proc->callref = skb->data[2]; /* cr flag */
	for (i = 0; i < GLOBALM_LEN; i++)
		if ((mt == globalmes_list_nt[i].primitive) &&
		    ((1 << proc->state) & globalmes_list_nt[i].state))
			break;
	if (i == GLOBALM_LEN) {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "\ndss1_nt global state %d mt %x unhandled\n",
				l3_state_name(proc->state), mt);
		}
		MsgHead(p, proc->callref, MT_STATUS);
		*p++ = IE_CAUSE;
		*p++ = 0x2;
		*p++ = 0x81;
		*p++ = 81 |0x80;	/* invalid cr */
		*p++ = 0x14;		/* CallState */
		*p++ = 0x1;
		*p++ = proc->state & 0x3f;
		l = p - tmp;
		if (!(skb = l3_alloc_skb(l)))
			return;
		memcpy(skb_put(skb, l), tmp, l);
		l3_msg(proc->st, DL_DATA | REQUEST, skb);
	} else {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "dss1_nt global %d mt %x",
				proc->state, mt);
		}
		globalmes_list_nt[i].rout(proc, mt, skb);
	}
}

static void
dss1up_nt(struct PStack *st, int pr, void *arg)
{
	int i, mt, cr, cause, callState;
	char *ptr;
	u_char *p;
	struct sk_buff *skb = arg;
	struct l3_process *proc;
	int tei;

	switch (pr) {
		case (DL_DATA | INDICATION):
		case (DL_UNIT_DATA | INDICATION):
			break;
		case (DL_ESTABLISH | CONFIRM):
			//l3_debug(st, "dss1up_nt establish confirm, sending l3_msg");
			l3_msg(st, pr, arg);
			return;
			break;
		case (DL_ESTABLISH | INDICATION):
			//l3_debug(st, "dss1up_nt establish indication, sending l3_msg");
			l3_msg(st, pr, arg);
			return;
			break;
		case (DL_RELEASE | INDICATION):
			//l3_debug(st, "dss1up_nt release indication, sending l3_msg");
			l3_msg(st, pr, arg);
			return;
			break;
		case (DL_RELEASE | CONFIRM):
			//l3_debug(st, "dss1up_nt release confirm, sending l3_msg");
			l3_msg(st, pr, arg);
			return;
			break;
                default:
                        printk(KERN_ERR "HiSax dss1up_nt unknown pr=%04x\n", pr);
                        return;
	}

#ifdef CONFIG_HISAX_MULTIPOINT
	/* we always use the first l2 protocol stack for data to l3 stacks */
	st = st->l1.hardware->d_st;
#endif

	/* get tei from frame */
	tei = skb->data[1] >> 1;
#warning remove debug
printk(KERN_DEBUG "\nDEBUG: FRAME FROM TEI %d\n\n", tei);

	if (skb->len < 3) {
		l3_debug(st, "dss1up_nt frame too short(%d)", skb->len);
		dev_kfree_skb(skb);
		return;
	}

	if (skb->data[0] != PROTO_DIS_EURO) {
		if (st->l3.debug & L3_DEB_PROTERR) {
			l3_debug(st, "dss1up_nt%sunexpected discriminator %x message len %d",
				 (pr == (DL_DATA | INDICATION)) ? " " : "(broadcast) ",
				 skb->data[0], skb->len);
		}
		dev_kfree_skb(skb);
		return;
	}
	cr = getcallref(skb->data);
	if (skb->len < ((skb->data[1] & 0x0f) + 3)) {
		l3_debug(st, "dss1up_nt frame too short(%d)", skb->len);
		dev_kfree_skb(skb);
		return;
	}
	mt = skb->data[skb->data[1] + 2];
	if (st->l3.debug & L3_DEB_STATE)
		l3_debug(st, "dss1up_nt cr %d", cr);
	if (cr == -2) {  /* wrong Callref */
		if (st->l3.debug & L3_DEB_WARN)
			l3_debug(st, "dss1up_nt wrong Callref");
		dev_kfree_skb(skb);
		return;
	} else if (cr == -1) {	/* Dummy Callref */
		if (mt == MT_FACILITY)
			if ((p = findie(skb->data, skb->len, IE_FACILITY, 0))) {
				l3dss1_nt_parse_facility(st, NULL, 
					(pr == (DL_DATA | INDICATION)) ? -1 : -2, p); 
				dev_kfree_skb(skb);
				return;  
			}
		if (st->l3.debug & L3_DEB_WARN)
			l3_debug(st, "dss1up_nt dummy Callref (no facility msg or ie)");
		dev_kfree_skb(skb);
		return;
	} else if ((((skb->data[1] & 0x0f) == 1) && (0==(cr & 0x7f))) ||
		(((skb->data[1] & 0x0f) == 2) && (0==(cr & 0x7fff)))) {	/* Global CallRef */
		if (st->l3.debug & L3_DEB_STATE)
			l3_debug(st, "dss1up_nt Global CallRef");
		nt_global_handler(st, mt, skb);
		dev_kfree_skb(skb);
		return;
	} else if (!(proc = getl3proc_nt(st, cr, tei))) {
		/* No transaction process exist, that means no call with
		 * this callreference is active
		 */
		if (mt == MT_SETUP) {
			/* Setup creates a new transaction process */
			if (skb->data[2] & 0x80) {
				/* Setup with wrong CREF flag */
				if (st->l3.debug & L3_DEB_STATE)
					l3_debug(st, "dss1up_nt wrong CRef flag");
				dev_kfree_skb(skb);
				return;
			}
			if (!(proc = dss1_nt_new_l3_process(st, cr, tei))) {
				/* May be to answer with RELEASE_COMPLETE and
				 * CAUSE 0x2f "Resource unavailable", but this
				 * need a new_l3_process too ... arghh
				 */
				dev_kfree_skb(skb);
				return;
			}
		} else if (mt == MT_STATUS) {
			cause = 0;
			if ((ptr = findie(skb->data, skb->len, IE_CAUSE, 0)) != NULL) {
				ptr++;
				if (*ptr++ == 2)
					ptr++;
				cause = *ptr & 0x7f;
			}
			callState = 0;
			if ((ptr = findie(skb->data, skb->len, IE_CALL_STATE, 0)) != NULL) {
				ptr++;
				if (*ptr++ == 2)
					ptr++;
				callState = *ptr;
			}
			/* ETS 300-104 part 2.4.1
			 * if setup has not been made and a message type
			 * MT_STATUS is received with call state == 0,
			 * we must send nothing
			 */
			if (callState != 0) {
				/* ETS 300-104 part 2.4.2
				 * if setup has not been made and a message type
				 * MT_STATUS is received with call state != 0,
				 * we must send MT_RELEASE_COMPLETE cause 101
				 */
				if ((proc = dss1_nt_new_l3_process(st, cr, tei))) {
					proc->para.cause = 101;
					l3dss1_nt_msg_release_complete(proc, 0, NULL);
				}
			}
			dev_kfree_skb(skb);
			return;
		} else if (mt == MT_RELEASE_COMPLETE) {
			dev_kfree_skb(skb);
			return;
		} else {
			/* ETS 300-104 part 2
			 * if setup has not been made and a message type
			 * (except MT_SETUP and RELEASE_COMPLETE) is received,
			 * we must send MT_RELEASE_COMPLETE cause 81 */
			dev_kfree_skb(skb);
			if ((proc = dss1_nt_new_l3_process(st, cr, tei))) {
				proc->para.cause = 81;
				l3dss1_nt_msg_release_complete(proc, 0, NULL);
			}
			return;
		}
	}
	if (l3dss1_nt_check_messagetype_validity(proc, mt, skb)) {
		dev_kfree_skb(skb);
		return;
	}
// there is no display from TE!
//	if ((p = findie(skb->data, skb->len, IE_DISPLAY, 0)) != NULL) 
//	  l3dss1_nt_deliver_display(proc, pr, p); /* Display IE included */
	for (i = 0; i < DATASLLEN; i++)
		if ((mt == datastatelist_nt[i].primitive) &&
		    ((1 << proc->state) & datastatelist_nt[i].state))
			break;
	if (i == DATASLLEN) {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "\ndss1up_nt%sstate %s mt %#x unhandled\n",
				(pr == (DL_DATA | INDICATION)) ? " " : "(broadcast) ",
				l3_state_name(proc->state), mt);
		}
		if ((MT_RELEASE_COMPLETE != mt) && (MT_RELEASE != mt)) {
			proc->para.cause = 101;
			l3dss1_nt_status_send(proc, pr, skb);
		}
	} else {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "dss1up_nt%sstate %s mt %x",
				(pr == (DL_DATA | INDICATION)) ? " " : "(broadcast) ",
				l3_state_name(proc->state), mt);
		}
#ifdef CONFIG_HISAX_SIGNALMODE
		/* fill proc->para.setup */
		get_setupextensions(proc, skb);
#endif
		datastatelist_nt[i].rout(proc, pr, skb);
	}
	dev_kfree_skb(skb);
	return;
}

static void
dss1down_nt(struct PStack *st, int pr, void *arg)
{
	int i, cr;
	struct l3_process *proc;
	struct Channel *chan;

	if ((DL_ESTABLISH | REQUEST) == pr) {
		l3_msg(st, pr, NULL);
		return;
	} else if (((CC_SETUP | REQUEST) == pr) || ((CC_RESUME | REQUEST) == pr)) {
		chan = arg;
		cr = newcallref();
		cr |= 0x80;
		if ((proc = dss1_nt_new_l3_process(st, cr, 0))) {
			proc->chan = chan;
			chan->proc = proc;
#ifdef CONFIG_HISAX_SIGNALMODE
			if (chan->signalmode)
				memcpy(&proc->para.signal, &chan->setup_signal, sizeof(signal_msg));
			else
#endif
			memcpy(&proc->para.setup, &chan->setup, sizeof(setup_parm));
			proc->callref = cr;
		}
	} else {
		proc = arg;
	}
	if (!proc) {
		printk(KERN_ERR "HiSax dss1down_nt without proc pr=%04x\n", pr);
		return;
	}

	if ( pr == (CC_TDSS1_IO | REQUEST)) {
		l3dss1_nt_io_timer(proc); /* timer expires */ 
		return;
	}  
//l3_debug(st,"DOWNSLLEN=%d, st=%d prim=%#x",DOWNSLLEN,downstatelist_nt[0].state,downstatelist_nt[0].primitive);
	for (i = 0; i < DOWNSLLEN; i++)
		if ((pr == downstatelist_nt[i].primitive) &&
		    ((1 << proc->state) & downstatelist_nt[i].state))
			break;
	if (i == DOWNSLLEN) {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "\ndss1down_nt state %s prim %#x unhandled\n",
				l3_state_name(proc->state), pr);
		}
	} else {
		if (st->l3.debug & L3_DEB_STATE) {
			l3_debug(st, "dss1down_nt state %s prim %#x",
				l3_state_name(proc->state), pr);
		}
		downstatelist_nt[i].rout(proc, pr, arg);
	}
}

static void
dss1man_nt(struct PStack *st, int pr, void *arg)
{
        int i;
        struct l3_process *proc = arg;

        if (!proc) {
                printk(KERN_ERR "HiSax dss1man_nt without proc pr=%04x\n", pr);
                return;
        }
        for (i = 0; i < MANSLLEN; i++)
                if ((pr == manstatelist_nt[i].primitive) &&
                    ((1 << proc->state) & manstatelist_nt[i].state))
                        break;
        if (i == MANSLLEN) {
                if (st->l3.debug & L3_DEB_STATE) {
                        l3_debug(st, "\ncr %d dss1man_nt state %s prim %#x unhandled\n",
                                proc->callref & 0x7f, l3_state_name(proc->state), pr);
                }
        } else {
                if (st->l3.debug & L3_DEB_STATE) {
                        l3_debug(st, "\ncr %d dss1man_nt state %s prim %#x\n",
                                proc->callref & 0x7f, l3_state_name(proc->state), pr);
                }
                manstatelist_nt[i].rout(proc, pr, arg);
        }
}

void
setstack_dss1_nt(struct PStack *st)
{
	char tmp[64];
	int i;

	st->lli.l4l3 = dss1down_nt;
	st->lli.l4l3_proto = l3dss1_nt_cmd_global;
	st->l2.l2l3 = dss1up_nt;
	st->l3.l3ml3 = dss1man_nt;
	st->l3.N303 = 1;
	st->prot.dss1_nt.last_invoke_id = 0;
	st->prot.dss1_nt.invoke_used[0] = 1; /* Bit 0 must always be set to 1 */
	i = 1;
	while (i < 32) 
		st->prot.dss1_nt.invoke_used[i++] = 0;   

	if (!(st->l3.global = kmalloc(sizeof(struct l3_process), GFP_ATOMIC))) {
		printk(KERN_ERR "HiSax can't get memory for dss1_nt global CR\n");
	} else {
		st->l3.global->state = 0;
		st->l3.global->callref = 0;
		st->l3.global->next = NULL;
		st->l3.global->debug = L3_DEB_WARN;
		st->l3.global->st = st;
		st->l3.global->N303 = 1;
		st->l3.global->prot.dss1_nt.invoke_id = 0; 

		L3InitTimer(st->l3.global, &st->l3.global->timer);
	}
	strcpy(tmp, dss1_nt_revision);
	printk(KERN_INFO "HiSax: DSS1 NT-MODE Rev. %s\n", HiSax_getrev(tmp));
}
