/* $Id: tei.c,v 0.0.0.1 2001/11/20 14:19:36 kai Exp $
 *
 * Author       Andreas Eversberg
 *              based on the tei.c from Jan den Ouden
 *		and Karsten Keil
 * 
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 *
 * For changes and modifications please read
 * ../../../Documentation/isdn/HiSax.cert
 *
 * Thanks to    Jan den Ouden
 *              Fritz Elfert
 *		Karsten Keil
 *
 */

#define __NO_VERSION__
#include "hisax.h"
#include "isdnl2.h"
#include <linux/init.h>
#include <linux/random.h>

const char *tei_nt_revision = "$Revision: 0.0.0.1 $";

#define ID_REQUEST	1
#define ID_ASSIGNED	2
#define ID_DENIED	3
#define ID_CHK_REQ	4
#define ID_CHK_RES	5
#define ID_REMOVE	6
#define ID_VERIFY	7

#define TEI_ENTITY_ID	0xf

static struct Fsm teintfsm;

void tei_nt_handler(struct PStack *st, u_char pr, struct sk_buff *skb);

enum {
	ST_TEI_IDASSIGNED,
	ST_TEI_IDCHECK,
	ST_TEI_IDLE,
};

#define TEI_STATE_COUNT (ST_TEI_IDLE+1)

static char *strTeiState[] =
{
	"ST_TEI_IDASSIGNED",
	"ST_TEI_IDCHECK",
	"ST_TEI_IDLE",
};

enum {
	EV_IDREQ,
	EV_TCHKEVENT,
	EV_CHKRES,
	EV_T201,
};

#define TEI_EVENT_COUNT (EV_T201+1)

static char *strTeiEvent[] =
{
	"EV_IDREQ",
	"EV_TCHKEVENT",
	"EV_CHKRES",
	"EV_T201",
};

static void
put_tei_nt_msg(struct PStack *st, u_char m_id, unsigned int ri, u_char tei)
{
	struct sk_buff *skb;
	u_char *bp;

	if (!(skb = alloc_skb(8, GFP_ATOMIC))) {
		printk(KERN_WARNING "HiSax: No skb for TEI NT-MODE manager\n");
		return;
	}
	bp = skb_put(skb, 3);
	bp[0] = (TEI_SAPI << 2) | 0x2;
	bp[1] = (GROUP_TEI << 1) | 0x1;
	bp[2] = UI;
	bp = skb_put(skb, 5);
	bp[0] = TEI_ENTITY_ID;
	bp[1] = ri >> 8;
	bp[2] = ri & 0xff;
	bp[3] = m_id;
	bp[4] = (tei << 1) | 1;
	st->l2.l2l1(st, PH_DATA | REQUEST, skb);
}

static void
tei_nt_id_dummy(struct FsmInst *fi, int event, void *arg)
{
}

static void
tei_nt_id_t201_tout(struct FsmInst *fi, int event, void *arg)
{
	struct PStack *st = fi->userdata;

	if (st->ma.tei_nt_chkres > 1) { /* too many responses received */
		if (st->ma.debug)
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"check event: too many responses for tei %d, removing tei",st->ma.tei_nt);
		st->l3.l3l2(st, MDL_REMOVE | REQUEST, (void *) (long) st->ma.tei_nt);
		put_tei_nt_msg(st, ID_REMOVE, 0, st->ma.tei_nt);
		put_tei_nt_msg(st, ID_REMOVE, 0, st->ma.tei_nt);
		st->ma.tei_nt=-1;
		FsmChangeState(fi, ST_TEI_IDLE);
	} else if (st->ma.tei_nt_chkres == 0) { /* no tei */
		if (st->ma.tei_nt_chkreq  == 1) { /* one request */
			if (st->ma.debug>1)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"check event: no response from tei %d, trying again",st->ma.tei_nt);
			st->ma.tei_nt_chkreq=2;
			st->ma.tei_nt_chkresfirst=0;
			put_tei_nt_msg(st, ID_CHK_REQ, 0, st->ma.tei_nt);
			FsmDelTimer(&st->ma.t201, 1);
			FsmAddTimer(&st->ma.t201, st->ma.T201, EV_T201, NULL, 3);
		} else { /* second request */
			if (st->ma.tei_nt_chkresfirst == 1) { /* no tei, but we got one at the first request */
				if (st->ma.debug>1)
					st->ma.tei_m.printdebug(&st->ma.tei_m,
						"check event: no response from tei %d, tei exists because already got response the first request",st->ma.tei_nt);
			} else {
				if (st->ma.debug)
					st->ma.tei_m.printdebug(&st->ma.tei_m,
						"check event: no response from tei %d, removing tei",st->ma.tei_nt);
				st->l3.l3l2(st, MDL_REMOVE | REQUEST, (void *) (long) st->ma.tei_nt);
				st->ma.tei_nt=-1;
				FsmChangeState(fi, ST_TEI_IDLE);
			}
		}
	} else { /* got one response */
		if (st->ma.tei_nt_chkreq  == 1) { /* one request */
			if (st->ma.debug>1)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"check event: response from tei %d, checking again",st->ma.tei_nt);
			st->ma.tei_nt_chkreq=2;
			st->ma.tei_nt_chkresfirst=1; /* first response */
			st->ma.tei_nt_chkres=0; /* prepare second response */
			put_tei_nt_msg(st, ID_CHK_REQ, 0, st->ma.tei_nt);
			FsmDelTimer(&st->ma.t201, 1);
			FsmAddTimer(&st->ma.t201, st->ma.T201, EV_T201, NULL, 3);
		} else { /* second request */
			if (st->ma.debug>1)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"check event: response from tei %d, tei exists",st->ma.tei_nt);
			FsmChangeState(fi, ST_TEI_IDASSIGNED);
			FsmDelTimer(&st->ma.tchkevent, 1);
			FsmAddTimer(&st->ma.tchkevent, st->ma.Tchkevent, EV_TCHKEVENT, NULL, 1);
		}
	}
}

/* there is a tei check response. lets see if this response has correct ri and correct tei */

static void
tei_nt_id_chkres(struct FsmInst *fi, int event, void *arg)
{
	struct PStack *st = fi->userdata;
	struct sk_buff *skb = arg;
	int tei;

	tei = skb->data[4] >> 1;

	if (tei==st->ma.tei_nt) { /* correct tei */
		if (st->ma.debug>1)
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"check event: got check response from tei %d",
				tei);
		st->ma.tei_nt_chkres++;
	} else {
		if (st->ma.debug)
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"check event: no request for tei %d (or maybe response too late)",
				tei);
	}
}

/* it is time to do a tei check */

static void
tei_nt_id_check(struct FsmInst *fi, int event, void *arg)
{
	struct PStack *st = fi->userdata;

	if (st->ma.debug>1)
		st->ma.tei_m.printdebug(&st->ma.tei_m,
			"check event: need to check tei %d",st->ma.tei_nt);
	put_tei_nt_msg(st, ID_CHK_REQ, 0, st->ma.tei_nt);
	FsmChangeState(fi, ST_TEI_IDCHECK);
	FsmDelTimer(&st->ma.t201, 1);
	FsmAddTimer(&st->ma.t201, st->ma.T201, EV_T201, NULL, 3);
	st->ma.tei_nt_chkreq=1;
	st->ma.tei_nt_chkres=0;
	st->ma.tei_nt_chkresfirst=0;
}

/* a tei has been requested and lets see if we can assign one */

static void
tei_nt_id_request(struct FsmInst *fi, int event, void *arg)
{
	struct PStack *st = fi->userdata;
	struct sk_buff *skb = arg;
	struct IsdnCardState *cs = st->l1.hardware;
	struct PStack *stptr = cs->stlist;
	int ri, tei;

	ri = ((unsigned int) skb->data[1] << 8) + skb->data[2];
	tei = skb->data[4] >> 1;

	if (tei == 127) { /* any tei */
		/* search for free tei */
		tei = 64;
		while(stptr) {
			if (stptr->ma.tei_m.state == ST_TEI_IDLE) {
				/* found unassigned stack */
				st = stptr; /* switch to correct stack */
				break;
			}
#warning remove debug
printk("\nDEBUG: tei=%d used, trying next!\n", stptr->ma.tei_nt);
			tei++;
			stptr = stptr->next;
		}
		if (!stptr) {
			if (st->ma.debug)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"request for tei denied because no more available");
			put_tei_nt_msg(st, ID_DENIED, ri, 127);
			return;
		}
	} else {
		/* check if given tei is available */
		int ltei = 64;
		while(stptr) {
			if (stptr->ma.tei_m.state==ST_TEI_IDLE && tei==ltei) {
				/* found unassigned stack */
				st = stptr; /* switch to correct stack */
				break;
			}
			ltei++;
			stptr = stptr->next;
		}
		if (!stptr) {
			if (st->ma.debug)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"request for tei %d denied because requested tei is already assigned or no more free stacks",
					tei);
			put_tei_nt_msg(st, ID_DENIED, ri, tei);
			return;
		}
	}
	/* assign TEI */
	if (st->ma.debug)
		st->ma.tei_m.printdebug(&st->ma.tei_m,
			"request for tei %d, assigning",
			tei);
	st->ma.tei_nt=tei;
	st->l3.l3l2(st, MDL_ASSIGN | REQUEST, (void *) (long) tei);
	put_tei_nt_msg(st, ID_ASSIGNED, ri, tei);
	FsmChangeState(fi, ST_TEI_IDASSIGNED);
	FsmDelTimer(&st->ma.tchkevent, 1);
	FsmAddTimer(&st->ma.tchkevent, st->ma.Tchkevent, EV_TCHKEVENT, NULL, 3);
}

static void
tei_nt_l1l2(struct PStack *st, int pr, void *arg)
{
	struct sk_buff *skb = arg;
	int mt;

	if (test_bit(FLG_FIXED_TEI, &st->l2.flag)) {
		dev_kfree_skb(skb);
		return;
	}

	if (pr == (PH_DATA | INDICATION)) {
		if (skb->len < 3) {
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"short mgr frame %ld/3", skb->len);
		} else if ((skb->data[0] != ((TEI_SAPI << 2) | 0)) ||
			   (skb->data[1] != ((GROUP_TEI << 1) | 1))) {
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"wrong mgr sapi/tei %x/%x",
				skb->data[0], skb->data[1]);
		} else if ((skb->data[2] & 0xef) != UI) {
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"mgr frame is not ui %x", skb->data[2]);
		} else {
			skb_pull(skb, 3);
			if (skb->len < 5) {
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"short mgr frame %ld/5", skb->len);
			} else if (skb->data[0] != TEI_ENTITY_ID) {
				/* wrong management entity identifier, ignore */
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"tei handler wrong entity id %x",
					skb->data[0]);
			} else {
				mt = skb->data[3];
				if (mt == ID_REQUEST) {
					/* stack will be searched during handling */
					FsmEvent(&st->ma.tei_m, EV_IDREQ, skb);
				}
				else if (mt == ID_CHK_RES) {
					/* handle event on all stacks */
					while(st) {
						FsmEvent(&st->ma.tei_m, EV_CHKRES, skb);
						st = st->next;
					}
				} else {
					st->ma.tei_m.printdebug(&st->ma.tei_m,
						"tei handler wrong mt %x\n", mt);
				}
			}
		}
	} else {
		st->ma.tei_m.printdebug(&st->ma.tei_m,
			"tei handler wrong pr %x\n", pr);
	}
	dev_kfree_skb(skb);
}

static void
tei_nt_l2tei(struct PStack *st, int pr, void *arg)
{
	struct IsdnCardState *cs;

	if (test_bit(FLG_FIXED_TEI, &st->l2.flag)) {
		if (pr == (MDL_ASSIGN | INDICATION)) {
			if (st->ma.debug)
				st->ma.tei_m.printdebug(&st->ma.tei_m,
					"fixed assign tei %d", st->l2.tei);
			st->l3.l3l2(st, MDL_ASSIGN | REQUEST, (void *) (long) st->l2.tei);
			cs = (struct IsdnCardState *) st->l1.hardware;
			cs->cardmsg(cs, MDL_ASSIGN | REQUEST, NULL);
		}
		return;
	}
	switch (pr) {
                case (MDL_ASSIGN | INDICATION):
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"unknown to l2 tei assign indication %d (channel should not be used if tei is not assigned!)\n", (long) arg);
//                        FsmEvent(&st->ma.tei_m, EV_IDREQ, arg);
                        break;
                case (MDL_ERROR | REQUEST):
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"l2 tei error requst %d\n", (long) arg);
			put_tei_nt_msg(st, ID_REMOVE, 0, st->ma.tei_nt);
			put_tei_nt_msg(st, ID_REMOVE, 0, st->ma.tei_nt);
			st->ma.tei_nt = -1;
			FsmChangeState(&(st->ma.tei_m), ST_TEI_IDLE);
                        break;
		default:
			st->ma.tei_m.printdebug(&st->ma.tei_m,
				"unknown to l2 tei requst %d\n", (long) arg);
			break;
	}
}

static void
tei_nt_debug(struct FsmInst *fi, char *fmt, ...)
{
	va_list args;
	struct PStack *st = fi->userdata;
	char title[64];

	sprintf(title, "Card%d TEI (nt-mode) ", ((struct IsdnCardState *)st->l1.hardware)->myid+1);

	va_start(args, fmt);
	VHiSax_putstatus(st->l1.hardware, title, fmt, args);
	va_end(args);
}

void
setstack_tei_nt(struct PStack *st)
{
	st->l2.l2tei = tei_nt_l2tei;
	st->ma.T201 = 2500;		/* T201  2500 milliseconds (see 5.9.6 of itu-t q.920) */
	st->ma.Tchkevent = 200000;	/* 200 seconds (3 minutes) */
	st->l1.l1tei = tei_nt_l1l2;
	st->ma.tei_nt = -1;		/* currently assigned by NT */
	st->ma.debug = 1; /* 0= off 1=brief 2=detailed */
	st->ma.tei_m.fsm = &teintfsm;
	st->ma.tei_m.state = ST_TEI_IDLE;
	st->ma.tei_m.debug = 0;
	st->ma.tei_m.userdata = st;
	st->ma.tei_m.userint = 0;
	st->ma.tei_m.printdebug = tei_nt_debug;
	FsmInitTimer(&st->ma.tei_m, &st->ma.tchkevent);
	FsmInitTimer(&st->ma.tei_m, &st->ma.t201);
}

void
init_tei_nt(struct IsdnCardState *cs, int protocol)
{
}

void
release_tei_nt(struct IsdnCardState *cs)
{
	struct PStack *st = cs->stlist;

	while (st) {
		FsmDelTimer(&st->ma.tchkevent, 1);
		FsmDelTimer(&st->ma.t201, 1);
		st = st->next;
	}
}

static struct FsmNode TeiNTFnList[] __initdata =
{
	{ST_TEI_IDLE, EV_IDREQ, tei_nt_id_request},
	{ST_TEI_IDLE, EV_TCHKEVENT, tei_nt_id_dummy},
	{ST_TEI_IDLE, EV_CHKRES, tei_nt_id_dummy},
	{ST_TEI_IDASSIGNED, EV_TCHKEVENT, tei_nt_id_check},
	{ST_TEI_IDCHECK, EV_IDREQ, tei_nt_id_request},
	{ST_TEI_IDCHECK, EV_CHKRES, tei_nt_id_chkres},
	{ST_TEI_IDCHECK, EV_T201, tei_nt_id_t201_tout},
};

#define TEI_NT_FN_COUNT (sizeof(TeiNTFnList)/sizeof(struct FsmNode))

int __init
TeiNTNew(void)
{
	teintfsm.state_count = TEI_STATE_COUNT;
	teintfsm.event_count = TEI_EVENT_COUNT;
	teintfsm.strEvent = strTeiEvent;
	teintfsm.strState = strTeiState;
	return FsmNew(&teintfsm, TeiNTFnList, TEI_NT_FN_COUNT);
}

void
TeiNTFree(void)
{
	FsmFree(&teintfsm);
}
